'''
Definition for the class Parameters. Used for storing the fitting parameters
in GenX.
Programmer: Matts Bjorck
'''

import numpy as np
import string
from genx.gui_logging import iprint

# ==============================================================================


class Parameters:
    """
    Class for storing the fitting parameters in GenX
    """
    # Parameters used for saving the object state
    export_parameters = {'data_labels': list, 'init_data': list, 'data': list}

    def __init__(self, model=None):
        self.data_labels = ['Parameter', 'Value', 'Fit', 'Min', 'Max', 'Error']
        self.init_data = ['', 0.0, False, 0.0, 0.0, 'None']
        self.data = [self.init_data[:]]
        self.model = model
        self.string_dtype = "S100"

    def write_h5group(self, group):
        """ Export the members in the object to a h5py group.
        :param group: h5py Group to export to
        :return:
        """
        group['data_labels'] = [label.encode('utf-8') for label in self.data_labels]
        #print np.array([r[0] for r in self.data], dtype='S50')
        group['data col 0'] = [r[0].encode('utf-8') for r in self.data]
        group['data col 1'] = [r[1] for r in self.data]
        group['data col 2'] = [r[2] for r in self.data]
        group['data col 3'] = [r[3] for r in self.data]
        group['data col 4'] = [r[4] for r in self.data]
        group['data col 5'] = [r[5].encode('utf-8') for r in self.data]

    def read_h5group(self, group):
        """ Import data to the object from a h5py group

        :param group: h5py Group to import from
        :return:
        """
        self.data_labels = [item.decode('utf-8') for item in list(group['data_labels'][()])]
        self.data = [[c0.decode('utf-8'), float(c1), bool(c2), float(c3), float(c4), c5.decode('utf-8')]
                     for (c0, c1, c2, c3, c4, c5) in
                     zip(group['data col 0'][()], group['data col 1'][()],
                         group['data col 2'][()], group['data col 3'][()],
                         group['data col 4'][()], group['data col 5'][()])]

    def to_dict(self):
        """Creates a dict from the names in the columns, returns a Dict"""
        d = {}
        for i in range(len(self.data_labels)):
            d[self.data_labels[i]] = [r[i] for r in self.data]
        return d

    def set_value(self, row, col, value):
        """ Set a value in the parameter grid """
        self.data[row][col] = value
    
    def get_value(self, row, col):
        """ Get the value in the grid """
        return self.data[row][col]

    def name_in_grid(self, name):
        """ Checks if name is a parameter in the grid

        :param name: string representation of a parameter
        :return:
        """
        par_names = [row[0] for row in self.data]
        return name in par_names

    def get_names(self):
        """ Returns the parameter names

        :return:
        """
        par_names = [row[0] for row in self.data]

        return par_names


    def get_value_by_name(self, name):
        """Get the value for parameter name. Returns None if name can not be found.

        :param name:
        :return: Value or None
        """
        par_names = [row[0] for row in self.data]
        if name in par_names:
            value = self.data[par_names.index(name)][1]
        else:
            value = None
        return value

    def get_fit_state_by_name(self, name):
        """Get the fitting state for parameter name. Returns None if name can not be found.

        :param name:
        :return: int 0 - not found, 1 - fitted, 2 - defined but constant
        """
        par_names = [row[0] for row in self.data]
        state = 0
        if name in par_names:
            if self.data[par_names.index(name)][2]:
                state = 1
            else:
                state = 2
        return state

    def set_fit_state_by_name(self, name, value, state, min_val, max_val):
        """ Set the fit state by name accoring to the following states
        0 -- remove from grid.
        1 -- add if not exist, set parameter to be fitted (True)
        2 -- add if not exist, set parameter to not be fitted (False)

        :param name: parameter name
        :param state: state, see above.
        :return:
        """
        par_names = [row[0] for row in self.data]
        if state == 0 and name in par_names:
            self.delete_rows([par_names.index(name),])
        elif state == 1 or state == 2:
            if state == 1:
                fit = True
            else:
                fit = False
            if name in par_names:
                self.data[par_names.index(name)][1] = value
                self.data[par_names.index(name)][2] = fit
            else:
                self.append()
                self.data[-1][0] = name
                self.data[-1][1] = value
                self.data[-1][2] = fit
                self.data[-1][3] = min_val
                self.data[-1][4] = max_val
                self.data[-1][5] = '-'


    def get_len_rows(self):
        return len(self.data)
    
    def get_len_cols(self):
        return len(self.data[0])
    
    def get_len_fit_pars(self):
        return sum([row[2] for row in self.data])
        
    def get_col_headers(self):
        return self.data_labels[:]
        
    def delete_rows(self, rows):
        ''' Delete the rows in the list rows ...'''
        delete_count = 0
        rows=rows[:]
        rows.sort()
        
        for i in rows:
            #Note index changes as we delete values. thats why rows has to be sorted
            try:
                self.data.pop(i - delete_count)
            except:
                pass
            else:
                delete_count += 1
                
        return delete_count
    
    def insert_row(self, row):
        ''' Insert a new row at row(int). '''
        self.data.insert(row, self.init_data[:])

    def move_row_up(self, row):
        """ Move row up.

        :param row: Move row up one row.
        :return: Boolean True if the row was moved, otherwise False.
        """

        if row != 0 and row < self.get_len_rows():
            self.data.insert(row - 1, self.data.pop(row))
            return True
        else:
            return False

    def move_row_down(self, row):
        """ Move row up.

        :param row: Move row down one row.
        :return: Boolean True if the row was moved, otherwise False.
        """

        if row < self.get_len_rows() - 1:
            self.data.insert(row + 1, self.data.pop(row))
            return True
        else:
            return False

    def _sort_key_func(self, item):
        class_name = ''
        pname = item[0]
        obj_name = item[0]
        if self.model is not None:
            if item[0].count('.') > 0 and self.model.compiled:
                pieces = item[0].split('.')
                obj_name = pieces[0]
                pname = pieces[1]
                obj = self.model.eval_in_model(obj_name)
                class_name = obj.__class__.__name__

        return string.lower(class_name), string.lower(pname), string.lower(obj_name)

    def sort_rows(self):
        """ Sort the rows in the table

        :return: Boolean to indicate success (True)
        """
        self.data.sort(key=self._sort_key_func)
        return True

    def append(self):
        self.data.append(self.init_data[:])
        
    def get_fit_pars(self):
        ''' Returns the variables needed for fitting '''
        #print 'Data in the parameters class: ', self.data
        rows = list(range(len(self.data)))
        row_nmb=[nmb for nmb in rows if self.data[nmb][2] and\
                not self.data[nmb][0]=='']
        funcs=[row[0] for row in self.data if row[2] and not row[0] == '']
        mytest=[row[1] for row in self.data if row[2] and not row[0] == '']
        min=[row[3] for row in self.data if row[2] and not row[0] == '']
        max=[row[4] for row in self.data if row[2] and not row[0] == '']
        return (row_nmb, funcs, mytest, min, max)
    
    def get_pos_from_row(self, row):
        '''get_pos_from_row(self) --> pos [int]
        
        Transform the row row to the position in the fit_pars list
        '''
        rows = list(range(row + 1))
        row_nmb=[nmb for nmb in rows if self.data[nmb][2] and\
                not self.data[nmb][0] == '']
        return len(row_nmb) - 1
    
    def get_sim_pars(self):
        ''' Returns the variables needed for simulation '''
        funcs = [row[0] for row in self.data if not row[0] == '']
        mytest = [row[1] for row in self.data if not row[0] == '']
        return (funcs, mytest)

    def get_sim_pos_from_row(self, row):
        '''Transform a row to a psoitions in the sim list 
        that is returned by get_sim_pars
        '''
        rows = list(range(row + 1))
        row_nmb=[nmb for nmb in rows if not self.data[nmb][0] == '']
        return len(row_nmb) - 1
       
    def set_value_pars(self, value):
        ''' Set the values of the parameters '''
        valueindex = 0
        for row in  self.data:
            if row[2] and not row[0] == '':
                row[1] = value[valueindex]
                valueindex = valueindex + 1
                
    def set_error_pars(self, value):
        ''' Set the errors on the parameters '''
        valueindex = 0
        for row in  self.data:
            if row[2] and not row[0] == '':
                row[5] = value[valueindex]
                valueindex = valueindex + 1

    def clear_error_pars(self):
        ''' clears the errors in the parameters'''
        for row in  self.data:
            row[5] = '-' 
                
    def set_data(self, data):
        rowi = 0
        coli = 0
        for row in data:
            for col in row:
                self.set_value(rowi, coli, col)
                coli = coli + 1
            rowi = rowi + 1
            coli = 0
            
    def get_data(self):
        return self.data[:]
    
    def get_ascii_output(self):
        '''get_ascii_output(self) --> text [string]
        
        Returns the parameters grid as an ascii string.
        '''
        text = '#'
        # Show the data labels but with a preceeding # to denote a comment
        for label in self.data_labels:
            text += label + '\t'
        text += '\n'
        for row in self.data:
            for item in row:
                # special handling of floats to reduce the 
                # col size use 5 significant digits
                if type(item) == type(10.0):
                    text += '%.4e\t'%item
                else:
                    text += item.__str__() + '\t'
            text += '\n'
        return text
    
    def _parse_ascii_input(self, text):
        '''parse_ascii_input(self, text) --> list table
        
        Parses an ascii string to a parameter table. returns a list table if 
        sucessful otherwise it returns None
        '''
        table = []
        sucess = True
        lines = text.split('\n')
        for line in lines[:-1]:
            # If line not commented
            if line[0] != '#' and line[0] != '\n':
                line_strs = line.split('\t')
                # Check the format is it valid?
                if len(line_strs) > 7 or len(line_strs) < 6:
                    if len(line_strs) == 1:
                        break
                    else:
                        sucess = False
                        break
                try:
                    par = line_strs[0]
                    val = float(line_strs[1])
                    fitted = line_strs[2].strip() == 'True'\
                                or line_strs[2].strip() == '1'
                    min = float(line_strs[3])
                    max = float(line_strs[4])
                    error = line_strs[5]
                except Exception as e:
                    sucess = False
                    break
                else:
                    table.append([par, val, fitted, min, max, error])
        if sucess:
            return table
        else:
            return None
        
    def set_ascii_input(self, text):
        '''set_ascii_input(self, text) --> None
        
        If possible parse the text source and set the current parameters table
        to the one given in text.
        '''
        table = self._parse_ascii_input(text)
        if table:
            self.data = table
            return True
        else:
            return False
    
    def safe_copy(self, object):
        '''safe_copy(self, object) --> None
        
        Does a safe copy from object into this object.
        '''
        self.data = object.data[:]
    
    def copy(self):
        '''get_copy(self) --> copy of Parameters
        
        Does a copy of the current object.
        '''
        new_pars = Parameters()
        new_pars.data = self.data[:]
        
        return new_pars

if __name__ == '__main__':
    p = Parameters()
    p.append()
    import h5py
    f = h5py.File('test.hdf', 'w')
    g = f.create_group('parameters')
    p.write_h5group(g)
    f.close()
    p2 = Parameters()
    f = h5py.File('test.hdf', 'r')
    p2.read_h5group(f['parameters'])
    iprint(p2.data)


''' A library for handling optical constants and scattering lengths
in an efficent way. Based on a base class database which is subclassed
for each case.
Programmer: Matts Bjorck
Last changed: 2009-03-19
'''

import numpy as np
import re, string
# Imports needed for the dispersive table implementation
from scipy import interpolate
from . import refl
from genx.gui_logging import iprint

#==============================================================================
class Func(object):
    '''A function object which stores the real function so it can be 
    dynamically replaced from its parents.
    '''
    def __init__(self, func):
        self.func = func
    
    def __call__(self, *args):
        return self.func(*args)
    
    def replace_func(self, func):
        self.func = func

class Proxy(object):
    '''Proxy class borrowed from ASPN Recipe 496741 at adress
    http://code.activestate.com/recipes/496741/
    '''
    __slots__ = ["_obj", "__weakref__"]
    def __init__(self, obj):
        object.__setattr__(self, "_obj", obj)
    
    #
    # proxying (special cases)
    #
    def __getattribute__(self, name):
        return getattr(object.__getattribute__(self, "_obj"), name)
    def __delattr__(self, name):
        delattr(object.__getattribute__(self, "_obj"), name)
    def __setattr__(self, name, value):
        setattr(object.__getattribute__(self, "_obj"), name, value)
    
    def __bool__(self):
        return bool(object.__getattribute__(self, "_obj"))
    def __str__(self):
        return str(object.__getattribute__(self, "_obj"))
    def __repr__(self):
        return repr(object.__getattribute__(self, "_obj"))
    def __call__(self, *args, **kwargs):
        return object.__getattribute__(self, "_obj").__call__(*args, **kwargs)
    
    def _change_object(self, obj):
        object.__setattr__(self, "_obj", obj)
    #
    # factories
    #
    _special_names = [
        '__abs__', '__add__', '__and__', '__cmp__', '__coerce__', 
        '__contains__', '__delitem__', '__delslice__', '__div__', '__divmod__', 
        '__eq__', '__float__', '__floordiv__', '__ge__', '__getitem__', 
        '__getslice__', '__gt__', '__hash__', '__hex__', '__iadd__', '__iand__',
        '__idiv__', '__idivmod__', '__ifloordiv__', '__ilshift__', '__imod__', 
        '__imul__', '__int__', '__invert__', '__ior__', '__ipow__', '__irshift__', 
        '__isub__', '__iter__', '__itruediv__', '__ixor__', '__le__', '__len__', 
        '__long__', '__lshift__', '__lt__', '__mod__', '__mul__', '__ne__', 
        '__neg__', '__oct__', '__or__', '__pos__', '__pow__', '__radd__', 
        '__rand__', '__rdiv__', '__rdivmod__', '__reduce__', '__reduce_ex__', 
        '__repr__', '__reversed__', '__rfloorfiv__', '__rlshift__', '__rmod__', 
        '__rmul__', '__ror__', '__rpow__', '__rrshift__', '__rshift__', '__rsub__', 
        '__rtruediv__', '__rxor__', '__setitem__', '__setslice__', '__sub__', 
        '__truediv__', '__xor__', 'next',
    ]
    
    #@classmethod #This is the new way of doing a classmethod, starting from python 2.4
    def _create_class_proxy(cls, theclass):
        """creates a proxy for the given class"""
        
        def make_method(name):
            def method(self, *args, **kw):
                return getattr(object.__getattribute__(self, "_obj"), name)(*args, **kw)
            return method
        
        namespace = {}
        for name in cls._special_names:
            if hasattr(theclass, name) and not hasattr(cls, name):
                namespace[name] = make_method(name)
        return type("%s(%s)" % (cls.__name__, theclass.__name__), (cls,), namespace)
    # And, this is the old way. Which I stck with for compatiblity reasons
    _create_class_proxy = classmethod(_create_class_proxy)
    
    def __new__(cls, obj, *args, **kwargs):
        """
        creates an proxy instance referencing `obj`. (obj, *args, **kwargs) are
        passed to this class' __init__, so deriving classes can define an 
        __init__ method of their own.
        note: _class_proxy_cache is unique per deriving class (each deriving
        class must hold its own cache)
        """
        try:
            cache = cls.__dict__["_class_proxy_cache"]
        except KeyError:
            cls._class_proxy_cache = cache = {}
        try:
            theclass = cache[obj.__class__]
        except KeyError:
            cache[obj.__class__] = theclass = cls._create_class_proxy(obj.__class__)
        ins = object.__new__(theclass)
        #theclass.__init__(ins, obj, *args, **kwargs)
        return ins
    
def change_proxyobject(proxy, obj):
    '''
    changes the current object of proxy to the obj.
    Note that evil things can happen if not care is taken so the objects are
    of the same class
    '''
    object.__setattr__(proxy, '_obj', obj)
    
#==============================================================================
class Database(object):
    '''A database class where object memebers are dynamically acessed
    and stored. I.e. the constants are only looked up when needed.
    '''
    def __init__(self):
        ''' __init__(self) --> None
        '''
        object.__setattr__(self, 'stored_values', {})
        
    def __getattribute__(self, name):
        '''__getattr__(self, name) --> object
        
        Looks up and returns the attribure name
        '''
        if name.startswith('__'):
            return object.__getattribute__(self, name)
        name = name.lower()
        stored_values = object.__getattribute__(self, 'stored_values')
        if name in stored_values:
            return stored_values[name]
        else:
            try:
                stored_values[name] = object.__getattribute__(self,\
                        'lookup_value')(name)
            except (LookupError, IOError) as e:
                raise LookupError('The name %s does not exist in the'\
                    'database'%name)
            return stored_values[name]
        
    def __setattr__(self, name, value):
        '''__setattr__(self, name) --> None
        
        Just overlaoding the setattribute so a object cant be set.
        '''
        raise TypeError('This object does not support assignements')
    
    def lookup_value(self, name):
        '''lookup_value(self, name) --> object
        
        Used to (externally) lookup a value in a database to be inserted in
        local one for this object.
        '''
        iprint('Looking up value')
        return 1
    
    def reset_database(self):
        '''reset_database(self) --> None
        
        Resets the internal database
        '''
        stored_values = object.__getattribute__(self, 'stored_values')
        stored_values = {}
    

#==============================================================================
class FormFactor(Database):
    ''' A database for the x-ray formfactor which includes the 
    anomulous part as well as the angle dependent part. The object will
    return a function of sin(theta)/lambda
    '''
    def __init__(self, wavelength, f_calc):
        '''__init__(self, wavelength, f_calc) --> None
        
        wavelength [float] the wavelength of the radiation in AA, f_calc a 
        function which calculates (lookup the scattering factor)
        '''
        Database.__init__(self)
        object.__setattr__(self, 'wavelength', wavelength)
        object.__setattr__(self, 'f_calc', f_calc)
        
    def set_wavelength(self, wavelength):
        '''set_wavelength(self, wavelength) --> None
        
        sets the wavelength of the database. This will reset the 
        all the values. I.e when called each value will be reloaded
        '''
        # First check so we actually change the wavelength ...
        if abs(wavelength\
            - object.__getattribute__(self, 'wavelength')) > 1e-10:
            object.__setattr__(self, 'wavelength', wavelength)
            #object.__getattribute__(self, 'reset_database')()
            stored_vals = object.__getattribute__(self, 'stored_values')
            #for key in stored_vals:
            #    f = object.__getattribute__(self, 'lookup_value')(key)
            #    change_proxyobject(stored_vals[key], f)
            # Removing a bug with proxy adding does not work properly
            for key in stored_vals:
                f = object.__getattribute__(self, 'lookup_value')(key)
                stored_vals[key] = f
        
    def __getattribute__(self, name):
        '''__getattribute__(self, name) --> value
        
        Finds and returns a attribute
        '''
        if name == 'set_wavelength' or name.startswith('__'):
            return object.__getattribute__(self, 'set_wavelength')
        else:
            #print 'getting value'
            return Database.__getattribute__(self, name)
        
    def lookup_value(self, name):
        '''lookup_value(self, name) --> f [function]
        
        looks up a value in the external database
        '''
        wl = object.__getattribute__(self, 'wavelength')
        #f = Proxy((object.__getattribute__(self, 'f_calc')(name, wl)))
        # Removing a bug with proxy adding does not work properly
        f = object.__getattribute__(self, 'f_calc')(name, wl)
        return f
    
#==============================================================================
class ScatteringLength(Database):
    ''' A simpler version of the database where everything is loaded at once
    good for simple and small databases. I.e the scatteringlength for
    a neutron which is largely wavelength independent.
    '''
    def __init__(self, values):
        '''__init__(self, values) --> None
        
        values is a dictonary of key value pairs that are used in the database.
        '''
        Database.__init__(self)
        object.__setattr__(self, 'stored_values', values)
        
    def lookup_value(self, name):
        raise LookupError('The element %s does not exist in the database'%name)
    
#==============================================================================
# Function to load databases and or values


def load_f0dabax(filename, create_rho = False):
    '''load_dabax(filename) --> f0 (dictonary)
    
    loads a dabax file with f0 values and return a dictonary with 
    f0(sin(theta)/lambda) for the element given by the key name.
    '''
    def create_f(a, b, c):
        def f(sin_over_lambda):
            '''f(sin_over_lambda) --> float
            
            The atomic form factor for x-rays. The non-dispertive part.
            '''
            return c + sum(a[:, np.newaxis]*np.exp(-b[:, np.newaxis]\
                *sin_over_lambda**2), 0)
        return f

    def create_rho(a, b, c):
        def rho(r, fp, wb, occ):
            return (sum(a[:, np.newaxis]/\
                    (8.0*(np.pi*(b[:, np.newaxis]/16/np.pi**2+wb))**1.5)\
                     *np.exp(-(r**2.0)/4.0/(b[:, np.newaxis]/16/np.pi**2+wb)),\
                        0) + (c + fp)/(8.0*(np.pi*wb)**1.5)*\
                    np.exp(-(r**2.0)/4.0/wb))*occ
        return rho
    
    f = open(filename)
    real_label = ''
    temp_dict = {}
    for line in f.readlines():
        # Get the label for each line
        if line[0]=='#':
            label=line[1]
            ret=line[1:-1]
        else:
            label='D'
            ret=line[:-1]
        # Gets the real label, atom name
        if label == 'S':
            real_label = ret.split()[-1]
        # The row contains data
        if label == 'D':
            temp_dict[real_label.lower()] = [float(x) for x in ret.split()]
    
    f0 = {}
    rho0 = {}
    for key in list(temp_dict.keys()):
        temp = temp_dict[key]
        # change the name to not have operators...
        if key[-1] == '-':
            key = key[:-1] + 'm'
        if key[-1] == '+':
            key = key[:-1] + 'p'
        f0[key] = create_f(np.array(temp[:4]), np.array(temp[5:]), temp[4])
        rho0[key] = create_rho(np.array(temp[:4]), np.array(temp[5:]), temp[4])
    if create_rho:
        return f0, rho0
    else:
        return f0

def create_fp_lookup(path):
    '''create_f_lookup(filename) --> lookup_func(name, wavelength)
    
    Creates a lookup function to lookup element names and returns a function
    that yields dispersive scattering factors at Q = 0. NOTE wavelengths in AA
    '''
    def lookup_func(name, wavelength):
        '''lookup_func(name, wavelength) --> fp = f1 - 1.0J*f2
        
        looksup the scattering factors for a given wavelength given in AA for
        a element (note that the databases does not support ions). 
        The data given is the dispersive part + f0 (non dispersive) at Q = 0.
        '''
        e, f1, f2 = np.loadtxt(path + '%s.nff'%name.lower(), skiprows = 1,\
                unpack = True)
        energy = 1239.842/wavelength*10
        if energy >= e[-2] or energy <= e[1]:
            raise ValueError('The energy/wavelength is outside the databse'\
                + 'range, the energy should be inside [%f,%f] '%(e[1],e[-2]))
        pos1 = np.argmin(abs(e - energy))
        # Is the energy point to the right or left of the current point
        # If it is ontop it doesn't matter since the interpolation will be exact at
        # the endpoints
        if (e[pos1] - energy) > 0: 
            pos2 = pos1 - 1
        else:
            pos2 = pos1 + 1
        # A quick linear interpolation:
        f1_e = (energy - e[pos1])*(f1[pos2] - f1[pos1])/(e[pos2] - e[pos1])\
                + f1[pos1]
        f2_e = (energy - e[pos1])*(f2[pos2] - f2[pos1])/(e[pos2] - e[pos1])\
                + f2[pos1]
        
        return f1_e - 1.0J*f2_e
    return lookup_func

def create_fpdisp_lookup(path):
    '''create_f_lookup(filename) --> lookup_func(name)

    Creates a lookup function to lookup element names and returns a function
    that yields a function of dispersive scattering factors f(E) at Q = 0. NOTE energy is in eV
    '''
    def create_dispersion_func(name):
        e, f1, f2 = np.loadtxt(path + '%s.nff'%name.lower(), skiprows = 1,\
                unpack = True)
        f1interp = interpolate.interp1d(e, f1, kind = 'linear')
        f2interp = interpolate.interp1d(e, f2, kind = 'linear')

        def f(energy):
            return f1interp(energy) - 1.0J*f2interp(energy)

        return refl.ReflFunction(f, (np.mean(e),), {}, id = 'f(E)')
    return create_dispersion_func

def create_f_lookup(lookup_fp, f0):
    '''create_f_lookup(lookup_fp, f0) --> lookup_func(name, wavelength)
    
    combines a f0 dictonary [dict] witht the lookup function for the
     anoumoulous scattering factors. Note that it is assumed that lookup_fp 
    is a function on the form lookup_fp(name, wavelength) --> f1 - 1.0Jf2
    where name is the name of the element!
    '''
    def lookup_func(name, wavelength):
        ''' Looks up the total angle dependent form factor 
        f = f0 + f1 + 1.0J*f2 of element name. This dispersive part
        is independent on the ionicity but f0 not.
        '''
        # Check if name corrspond to an ion
        element = None
        if len(name) > 1:
            if name[-2].isdigit():
                element = name[:-2]
            else:
                element = name
        else:
            element = name
        # Remove the non dispersive part for the given element
        fp = lookup_fp(element, wavelength) - f0[element](0)
        f0_name = f0[name]
        def f(sin_over_lambda):
            return f0_name(sin_over_lambda) + fp
        
        return f
    
    return lookup_func

def create_rho_lookup(lookup_fp, rho0, f0):
    '''create_rho_lookup(lookup_fp, rho0, f0) --> lookup_func(name, wavelength)
    
    combines a rho0 dictonary [dict] with the lookup function for the
     anoumoulous scattering factors. Note that it is assumed that lookup_fp 
    is a function on the form lookup_fp(name, wavelength) --> f1 - 1.0Jf2
    where name is the name of the element!
    '''
    def lookup_func(name, wavelength):
        ''' Looks up the "electron density" of an atom as described by
        its formfactor, f = f0 + f1 + 1.0J*f2, of element name.
        The dispersive part is independent on the ionicity but f0 not.
        '''
        # Check if name corrspond to an ion
        element = None
        if len(name) > 1:
            if name[-2].isdigit():
                element = name[:-2]
            else:
                element = name
        else:
            element = name
        # Remove the non dispersive part for the given element
        fp = lookup_fp(element, wavelength) - f0[element](0)
        rho0_name = rho0[name]
        def rho(r, wb, occ):
            return rho0_name(r, fp, wb, occ) 
        
        return rho
    
    return lookup_func

def create_fw_lookup(lookup_fp, weight):
    '''create_fw_lookup(lookup_fp, weight) --> lookup_func(name, wavelength)
    
    combines a f0 dictonary [dict] witht the lookup function for the
     anoumoulous scattering factors. Note that it is assumed that lookup_fp 
    is a function on the form lookup_fp(name, wavelength) --> f1 - 1.0Jf2
    where name is the name of the element!
    '''
    def lookup_func(name, wavelength):
        ''' Looks up the total form factor at Q = 0
        f = f1 + 1.0J*f2 of element name. 
        '''
        fp = lookup_fp(name, wavelength)
        w = weight[name]
        
        return fp/(w/0.6022141)
    
    return lookup_func

def load_bdabax(filename):
    '''load_bdabax(filename) --> b (dictonary)
    
    loads a dabax file with b (sld for neutrons) values 
    and return a dictonary with b for the elements and isotopes 
    given by the key name.
    '''
    temp_dict = read_dabax(filename)
    # We are intrested in the scatttering lengths
    b_c = {}
    for key in temp_dict:
        val = temp_dict[key][1] + temp_dict[key][2]*1.0J
        if key[0].isdigit():
            b_c['i' + key] = val
        else:
            b_c[key] = val
            
    return b_c

def load_fdabax(filename):
    ''' loads a dabax file with the scattering length tables returns a lookup 
    function so that the wavelength can be changed. '''
    table = read_dabax(filename)
    def lookup_func(name, wavelength):
        ''' Looks up the total form factor at Q = 0
        f = f1 + 1.0J*f2 of element name. 
        '''
        el = np.array(table[name])
        el = el.reshape(len(el)/7, 7)
        e = el[:,0]
        f1 = el[:,1] 
        f2 = el[:,2]
        energy = 1239.842/wavelength*10/1e3 #keV
        if energy >= e[-2] or energy <= e[1]:
            raise ValueError('The energy/wavelength is outside the databse'\
                + 'range, the energy should be inside [%f,%f] '%(e[1],e[-2]))
        pos1 = np.argmin(abs(e - energy))
        # Is the energy point to the right or left of the current point
        # If it is ontop it doesn't matter since the interpolation will be exact at
        # the endpoints
        if (e[pos1] - energy) > 0: 
            pos2 = pos1 - 1
        else:
            pos2 = pos1 + 1
        # A quick linear interpolation:
        f1_e = (energy - e[pos1])*(f1[pos2] - f1[pos1])/(e[pos2] - e[pos1])\
                + f1[pos1]
        f2_e = (energy - e[pos1])*(f2[pos2] - f2[pos1])/(e[pos2] - e[pos1])\
                + f2[pos1]
        
        return f1_e - 1.0J*f2_e
    
    return lookup_func
    

def read_dabax(filename):
    '''read_dabax_dict(filename) --> temp_dict [dict]
    
    Loads an entire dabax file to a dictonary. Scan names are the 
    keys.
    '''
    def tofloat(x):
        try:
            f = float(x)
        except:
            f = np.nan
        return f
    f = open(filename)
    real_label = ''
    temp_dict = {}
    for line in f.readlines():
        # Get the label for each line
        if line[0] == '#':
            label = line[1]
            ret=line[1:-1]
        else:
            label = 'Data'
            ret = line[:-1]
        # Gets the real label, atom name
        if label == 'S':
            real_label = ret.split()[-1]
        # The row contains data
        if label == 'Data':
            # To get all values in the table
            if real_label.lower() not in temp_dict:
                temp_dict[real_label.lower()] = [tofloat(x.split('(')[0]) for x in ret.split()]
            else:
                temp_dict[real_label.lower()] += [tofloat(x.split('(')[0]) for x in ret.split()]
                
    return temp_dict

def load_atomic_weights_dabax(filename):
    '''load__atomic_weights_dabax(filename) --> w (dictonary)
    
    loads a dabax file with w (sld for neutrons) values 
    and return one dictonary with the atomic weights for the elements 
    by the key name. Note that all isotopes are includes with the an
    i in front of them.
    '''
    
    temp_dict = read_dabax(filename)
    # We are intrested in the scatttering lengths
    w_mean = {}
    for key in temp_dict:
        if key[0].isdigit():
            # This is an isotope
            w_mean['i' + key] = temp_dict[key][0]
            if len(temp_dict[key]) == 3:
                # The mean atomic mass for the element
                w_mean[key.lstrip(string.digits)] = temp_dict[key][2]
            
    return w_mean

def create_scatt_weight(scatt_dict, w_dict):
    '''create_bw(scatt_dict, w_dict) --> sw_dict
    
    Makes a scattering length database for using with densities (g/cm3)
    '''
    sw_dict = {}
    for key in scatt_dict:
        #print key, ' ',scatt_dict[key], ' ',w_dict[key]
        if key in w_dict:
            sw_dict[key] = scatt_dict[key]/complex(w_dict[key]/0.6022141)
    return sw_dict



'''
  Module used to setup the default GUI logging and messaging system.
  The system contains on a python logging based approach with logfile,
  console output and GUI output dependent on startup options and
  message logLevel.
'''

import sys
import atexit
import logging
import traceback
import inspect
import warnings
from numpy import seterr, seterrcall, ComplexWarning
from io import StringIO
from .version import __version__ as str_version

# default options used if nothing is set in the configuration
CONSOLE_LEVEL, FILE_LEVEL, GUI_LEVEL=logging.WARNING, logging.DEBUG, logging.INFO

# set log levels according to options
if 'pdb' in list(sys.modules.keys()) or 'pydevd' in list(sys.modules.keys()):
  # if common debugger modules have been loaded, assume a debug run
  CONSOLE_LEVEL, FILE_LEVEL, GUI_LEVEL=logging.INFO, logging.DEBUG, logging.INFO
elif '--debug' in sys.argv:
  sys.argv.remove('--debug')
  CONSOLE_LEVEL, FILE_LEVEL, GUI_LEVEL=logging.DEBUG, logging.DEBUG, logging.INFO

def excepthook_overwrite(*exc_info):
  logging.critical('python error', exc_info=exc_info)

def ip_excepthook_overwrite(self, etype, value, tb, tb_offset=None):
  logging.critical('python error', exc_info=(etype, value, tb))

def goodby():
  logging.info('*** GenX %s Logging ended ***'%str_version)

def iprint(*objects, sep=None, end=None, file=None, flush=False):
  """
  A logging function that behaves like print but uses logging.info.
  """
  if sep is None:
    sep=' '
  if end is None:
    end='\n'
  logging.info(sep.join(map(str, objects))+end)
  

class NumpyLogger(logging.getLoggerClass()):
  '''
    A logger that makes sure the actual function definition filename, lineno and function name
    is used for logging numpy floating point errors, not the numpy_logger function.
  '''

  if sys.version_info[0:2]>=(3, 2): #sinfo was introduced in python 3.2
    def makeRecord(self, name, lvl, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
      curframe=inspect.currentframe()
      calframes=inspect.getouterframes(curframe, 2)
      # stack starts with:
      # (this method, debug call, debug call rootlogger, numpy_logger, actual function, ...)
      ignore, fname, lineno, func, ignore, ignore=calframes[4]
      return logging.getLoggerClass().makeRecord(self, name, lvl, fname, lineno,
                                   msg, args, exc_info, func=func, extra=extra, sinfo=sinfo)
  else:
    def makeRecord(self, name, lvl, fn, lno, msg, args, exc_info, func=None, extra=None):
      curframe=inspect.currentframe()
      calframes=inspect.getouterframes(curframe, 2)
      # stack starts with:
      # (this method, debug call, debug call rootlogger, numpy_logger, actual function, ...)
      ignore, fname, lineno, func, ignore, ignore=calframes[4]
      return logging.getLoggerClass().makeRecord(self, name, lvl, fname, lineno,
                                                 msg, args, exc_info, func=func, extra=extra)

nplogger=None
def numpy_logger(err, flag):
  nplogger.debug('numpy floating point error encountered (%s)'%err)

def setup_system():
  logger=logging.getLogger()
  logger.setLevel(min(FILE_LEVEL, CONSOLE_LEVEL, GUI_LEVEL))

  # no console logger for windows (win32gui)
  console=logging.StreamHandler(sys.__stdout__)
  formatter=logging.Formatter('%(levelname) 7s: %(message)s')
  console.setFormatter(formatter)
  console.setLevel(CONSOLE_LEVEL)
  logger.addHandler(console)

  logging.getLogger('matplotlib').setLevel(logging.WARNING)
  if min(FILE_LEVEL, CONSOLE_LEVEL, GUI_LEVEL)>logging.DEBUG:
    logging.getLogger('numba').setLevel(logging.WARNING)
  logging.info('*** GenX %s Logging started ***'%str_version)

  # define numpy warning behavior
  global nplogger
  old_class=logging.getLoggerClass()
  logging.setLoggerClass(NumpyLogger)
  nplogger=logging.getLogger('numpy')
  nplogger.setLevel(logging.DEBUG)
  null_handler=logging.StreamHandler(StringIO())
  null_handler.setLevel(logging.CRITICAL)
  nplogger.addHandler(null_handler)
  logging.setLoggerClass(old_class)
  seterr(divide='call', over='call', under='ignore', invalid='call')
  #warnings.filterwarnings(action="error", category=ComplexWarning)
  logging.captureWarnings(True)
  seterrcall(numpy_logger)

  # write information on program exit
  # sys.excepthook=excepthook_overwrite
  atexit.register(goodby)

def activate_logging(logfile):
  logger=logging.getLogger()
  logfile=logging.FileHandler(logfile, 'w')
  formatter=logging.Formatter('[%(levelname)s] - %(asctime)s - %(filename)s:%(lineno)i:%(funcName)s %(message)s', '')
  logfile.setFormatter(formatter)
  logfile.setLevel(FILE_LEVEL)
  logger.addHandler(logfile)
  logger.info('*** GenX %s Logging started to file ***'%str_version)





''' Module to provide the Layer - Stack - Sample classes to build a sample for reflectivity modelling.

Classes:
ReflFunction - A function class that can be used as a parameter in the other classes.
is_reflfunction - Funtion that checks if an object belongs to the class ReflFunction
ReflBase - Base class for all the physical classes.
LayerBase - Base Layer class.
StackBase - Base Stack class.
SampleBase - Base Sample class.
InstrumentBase - Base Instrument class.
ModelFactory - A factory class that creates specific instances of the Base classes for a certain set of parameters.
MakeClasses - Function that creates the classes (backward comparability).
'''

import numpy as np
from genx.gui_logging import iprint


class ReflFunction:
    def __init__(self, function, validation_args, validation_kwargs, id=None):
        """ Creates the Refl Function given function. The arguments validation_args and
        validation_kwargs will be used to the validate the returned type of the function by passing
        them to function. The variable id should be a unique string to identify the type ReflFunction.
        """
        self.__func__ = function
        self.validation_args = validation_args
        self.validation_kwargs = validation_kwargs
        self.id = id

    def __call__(self, *args, **kwargs):
        return self.__func__(*args, **kwargs)

    def validate(self):
        ''' Function to test that the function returns the anticipated type
        '''
        return self.__call__(*self.validation_args, **self.validation_kwargs)

    def _check_obj(self, other):
        """ Checks the object other so that it fulfills the demands for arithmetic operations.
        """
        supported_types = [int, float, int, complex, np.float64, np.float32]
        if is_reflfunction(other):
            if self.id != other.id:
                raise TypeError("Two ReflFunction objects must have identical id's to conduct arithmetic operations")
        elif not type(other) in supported_types:
            raise TypeError("%s is not supported for arithmetic operations "% (repr(type(other))) +
                            "of a ReflFunction. It has to int, float, long or complex" )

    def __mul__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)*other(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)*other
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __rmul__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return other(*args, **kwargs)*self(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return other*self(*args, **kwargs)
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __add__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return self(*args, **kwargs) + other(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return self(*args, **kwargs) + other
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __radd__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return other(*args, **kwargs) + self(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return other + self(*args, **kwargs)
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __sub__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return self(*args, **kwargs) - other(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return self(*args, **kwargs) - other
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __rsub__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return other(*args, **kwargs) - self(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return other - self(*args, **kwargs)
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __div__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)/other(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)/other
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __rdiv__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return other(*args, **kwargs)/self(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return other/self(*args, **kwargs)
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __neg__(self):
        def new_func(*args, **kwargs):
            return -self(*args, **kwargs)
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __pos__(self):
        def new_func(*args, **kwargs):
            return self(*args, **kwargs)
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __pow__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)**other(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)**other
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __rpow__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return other(*args, **kwargs)**self(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return other**self(*args, **kwargs)
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)


def cast_to_array(list_of_obj, *args, **kwargs):
    ''' Casts an list_of_obj, can be a number or an ReflFunction, into an list of evaluated values'''
    id = ''
    shape = False
    ret_list = []
    for obj in list_of_obj:
        if is_reflfunction(obj):
            if id == '':
                id = obj.id
                ret_list.append(obj(*args, **kwargs))
                # Check if we have got an array
                if not np.isscalar(ret_list[-1]):
                    shape = ret_list[-1].shape
            elif id == obj.id:
                ret_list.append(obj(*args, **kwargs))
            else:
                TypeError("Two ReflFunction objects must have identical id's in order to merge them into an array")
        else:
            # We assume that this is an object that can be transformed into an array later on.
            ret_list.append(obj)
    # if we have an array make sure that all the objects have the same shape
    if shape:
        base_array = np.ones(shape)
        nret_list = []
        for item in ret_list:
            if np.isscalar(item):
                nret_list.append(item*base_array)
            else:
                nret_list.append(item)
        ret_list = nret_list

    return np.array(ret_list)

def harm_sizes(ar, shape, dtype = np.float64):
    '''Utility function to add an additional axis if needed to fulfill the size in shape'''
    ar = np.array(ar, dtype = dtype)
    if shape is None:
        return ar
    elif len(ar.shape) < len(shape):
        return np.array(ar[..., np.newaxis]*np.ones(shape), dtype = dtype)
    elif ar.shape == shape:
        return ar
    else:
        raise TypeError('The size of the array, %s, can not be changed to shape %s'%(ar.shape, shape))



def is_reflfunction(obj):
    ''' Convenience function to determine whether obj belongs to the ReflFunction class.
    Return boolean.
    '''
    return obj.__class__.__name__ == 'ReflFunction'

def resolve_par(obj, par_name, max_recur=100):
    """ Resolves a layer parameter and takes into account a coupling to another object.
    If obj.par_name is of type obj the relation will be recursive, obj.par_name.par_name.
    A maximum of 100 levels are conducted before an error is thrown.

    :param obj: An object
    :param par_name: A name of a member
    :param max_recur: Maximum number of recursions.
    :return: a value
    """
    val = getattr(obj, par_name)
    i = 0
    while type(val) is type(obj) and i < max_recur and not is_reflfunction(val):
        val = getattr(val, par_name)
        i += 1

    return val




class ReflBase:
    # _parameters is a dict of parameter names with their default values
    _parameters = {}

    def __init__(self, **kargs):
        # Setup all the parameters in the class
        for par in self._parameters:
            setattr(self, par, self._parameters[par])
            iscomplex = type(self._parameters[par]) is complex
            # Making the set function
            self._make_set_func(par, iscomplex)
            # Creating the get function
            self._make_get_func(par, iscomplex)

        # Set all parameters given as keyword arguments
        for k in kargs:
            if not k in self._parameters and not k in dir(self):
                continue
                raise ValueError('%s is not an parameter in %s' %
                                 (k, self.__class__))
            else:
                setattr(self, k, kargs[k])

    def _make_get_func(self, par, iscomplex=False):
        ''' Creates a get function for parameter par and binds it to the object
        '''

        def get_func():
            return getattr(self, par)

        get_func.__name__ = 'get' + par.capitalize()
        setattr(self, get_func.__name__, get_func)

        if iscomplex:
            def get_real_func():
                return getattr(self, par).real

            get_real_func.__name__ = get_func.__name__ + 'real'
            setattr(self, get_real_func.__name__, get_real_func)

            def get_imag_func():
                return getattr(self, par).imag

            get_imag_func.__name__ = get_func.__name__ + 'imag'
            setattr(self, get_imag_func.__name__, get_imag_func)

    def _make_set_func(self, par, iscomplex=False):
        ''' Creates a set function for parameter par and binds it to the object
            '''

        def set_func(val):
            setattr(self, par, val)

        set_func.__name__ = 'set' + par.capitalize()
        setattr(self, set_func.__name__, set_func)

        if iscomplex:
            def set_real_func(val):
                setattr(self, par, val + getattr(self, par).imag * 1J)

            set_real_func.__name__ = set_func.__name__ + 'real'
            setattr(self, set_real_func.__name__, set_real_func)

            def set_imag_func(val):
                setattr(self, par, val * 1J + getattr(self, par).real)

            set_imag_func.__name__ = set_func.__name__ + 'imag'
            setattr(self, set_imag_func.__name__, set_imag_func)


    def get_parameters(self):
        ''' Returns all the parameters of the current object '''

        return self.parameters


class InstrumentBase(ReflBase):
    _parameters = {}

    def __repr__(self):
        s = 'Instrument('

        for k in self._parameters:
            # if the type is a string...
            if type(getattr(self, k)) == type(''):
                stemp = "%s='%s', " % (k, str(getattr(self, k)))
            else:
                stemp = '%s=%s, ' % (k, str(getattr(self, k)))
            s = s + stemp
        return s[:-1] + ')'


class LayerBase(ReflBase):
    _parameters = {}

    def __repr__(self):
        s = 'Layer('
        for k in self._parameters:
            s += '%s=%s, ' % (k, str(getattr(self, k)))
        return s[:-2] + ')'


class StackBase(ReflBase):
    _parameters = {'Repetitions': 1}
    Layers = []
    Repetitions = 1

    def __init__(self, **kargs):
        ReflBase.__init__(self, **kargs)
        if 'Layers' in kargs:
            self.Layers = kargs['Layers']
            kargs.pop('Layers')
        else:
            self.Layers = []


    def __repr__(self):
        s = 'Stack: '
        for k in self._parameters:
            s += '%s=%s, ' % (k, str(getattr(self, k)))
        s = s[:-2] + '\n'
        it = len(self.Layers)
        for lay in range(it - 1, -1, -1):
            s += '\t' + repr(self.Layers[lay]) + '\n'
        return s

    def resolveLayerParameter(self, parameter):
        par = [resolve_par(lay, parameter)
               for lay in self.Layers] * int(self.Repetitions)
        return par


# Class Sample

class SampleBase(ReflBase):
    _parameters = {}
    Ambient = None
    Substrate = None
    Stacks = []

    def __init__(self, **kargs):
        ReflBase.__init__(self, **kargs)
        special_objects = ['Ambient', 'Substrate', 'Stacks']
        for p in special_objects:
            if p in kargs:
                setattr(self, p, kargs[p])
            else:
                raise ValueError('%s has to be defined to create a Sample object' % p)
        # Create a new dictonary for the new arguments
        nkargs = {}
        for p in kargs:
            if not p in special_objects:
                nkargs[p] = kargs[p]


    def __repr__(self):
        Add = 'Sample: '
        for k in self._parameters:
            if type(getattr(self, k)) == type(''):
                Add += "%s='%s', " % (k, str(getattr(self, k)))
            else:
                Add += '%s=%s, ' % (k, str(getattr(self, k)))
            #Add += '%s=%s, ' % (k, getattr(self, k).__repr__())
        if len(self._parameters) > 0:
            Add = Add[:-2] + '\n'
        else:
            Add = Add[:-1] + '\n'
        temp = [repr(x) for x in self.Stacks]
        temp.reverse()
        return (Add + 'Ambient:\n\t' + repr(self.Ambient) + '\n' + ''.join(temp)
                + 'Substrate:\n\t' + repr(self.Substrate))

    def resolveLayerParameters(self):
        par = self.Substrate._parameters.copy()
        for k in par:
            par[k] = [resolve_par(self.Substrate, k)]
        for k in self.Substrate._parameters:
            for stack in self.Stacks:
                par[k] = par[k] + stack.resolveLayerParameter(k)
            par[k] = par[k] + [resolve_par(self.Ambient, k)]
        return par


# Factory needs to:
#            add members and their names in the _parameter list
#            Add set and get methods for each parameter
class ModelFactory:
    ''' Class that creates new classes to implement certain parameters 
        dynamically.
    '''



    def __init__(self):
        class Layer(LayerBase): _parameters = LayerBase._parameters.copy()
        self.Layer = Layer

        class Stack(StackBase): _parameters = StackBase._parameters.copy()
        self.Stack = Stack

        class Sample(SampleBase): _parameters = SampleBase._parameters.copy()
        self.Sample = Sample

        class Instrument(InstrumentBase): _parameters = InstrumentBase._parameters.copy()
        self.Instrument = Instrument

    @staticmethod
    def _add_parameter(cls, par_name, def_val):
        cls._parameters.update({par_name: def_val})

    @staticmethod
    def _add_sim_method(cls, name, func):
        def method(self, *args):
            nargs = args[:-1] + (self, ) + (args[-1], )
            return func(*nargs)

        method.__name__ = 'Sim' + name
        setattr(cls, method.__name__, method)

    def set_layer_parameters(self, parameters):
        '''Adds extra parameters (dict with name: value entries) to the
            Layer class'''
        [self._add_parameter(self.Layer, k, parameters[k]) for k in parameters]

    def set_stack_parameters(self, parameters):
        '''Adds extra parameters (dict with name: value entries) to the
            Stack class'''
        # The if not is a hook to so that everything works with the old implementation
        [self._add_parameter(self.Stack, k, parameters[k]) for k in parameters
         if not k in ['Layers', 'Repetitions']]

    def set_sample_parameters(self, parameters):
        '''Adds extra parameters (dict with name: value entries) to the
            Sample class'''
        # The if not is a hook to so that everything works with the old implementation
        [self._add_parameter(self.Sample, k, parameters[k]) for k in parameters
         if not k in ['Ambient', 'Substrate', 'Stacks']]

    def set_simulation_functions(self, functions):
        ''' Adds the simulation functions to the sample class
        '''
        [self._add_sim_method(self.Sample, k, functions[k]) for k in functions]

    def set_instrument_parameters(self, parameters):
        '''Adds extra parameters (dict with name: value entries) to the
            Instrument class'''
        [self._add_parameter(self.Instrument, k, parameters[k]) for k in parameters]

    def get_layer_class(self):
        return self.Layer

    def get_stack_class(self):
        return self.Stack

    def get_sample_class(self):
        return self.Sample

    def get_instrument_class(self):
        return self.Instrument


# Function! to create classes with the model given by the input parameters
def MakeClasses(InstrumentParameters={'Wavelength': 1.54, 'Coordinates': 1}, \
                LayerParameters={'d': 0.0, 'sigma': 0.0, 'n': 0.0 + 0.0j}, \
                StackParameters={}, \
                SampleParameters={}, \
                SimulationFunctions={'Specular': lambda x: x, 'OffSpecular': lambda x: x}, \
                ModelID='Standard'):
    factory = ModelFactory()
    factory.set_layer_parameters(LayerParameters)
    factory.set_stack_parameters(StackParameters)
    factory.set_sample_parameters(SampleParameters)
    factory.set_simulation_functions(SimulationFunctions)
    factory.set_instrument_parameters(InstrumentParameters)
    Layer = factory.get_layer_class()
    Stack = factory.get_stack_class()
    Sample = factory.get_sample_class()
    Instrument = factory.get_instrument_class()
    return (Instrument, Layer, Stack, Sample)


if __name__ == '__main__':
    InstrumentParameters = {'probe': 'x-ray', 'wavelength': 1.54, 'coords': 'tth',
                            'I0': 1.0, 'res': 0.001,
                            'restype': 'no conv', 'respoints': 5, 'resintrange': 2, 'beamw': 0.01,
                            'footype': 'no corr', 'samplelen': 10.0, 'incangle': 0.0, 'pol': 'uu',
                            'Ibkg': 0.0, 'tthoff': 0.0}

    LayerParameters = {'sigma': 0.0, 'dens': 1.0, 'd': 0.0, 'f': (1.0 + 1.0j) * 1e-20,
                       'b': 0.0 + 0.0J, 'xs_ai': 0.0, 'magn': 0.0, 'magn_ang': 0.0}
    StackParameters = {'Layers': [], 'Repetitions': 1}
    SampleParameters = {'Stacks': [], 'Ambient': None, 'Substrate': None}

    (Instrument, Layer, Stack, Sample) = MakeClasses(InstrumentParameters, \
                                                     LayerParameters, StackParameters, SampleParameters,
                                                     {'Specualr': lambda x: x}, \
                                                     'test')
    inst = Instrument(footype='gauss beam', probe='x-ray', beamw=0.04, resintrange=2, pol='uu', wavelength=1.54,
                      respoints=5, Ibkg=0.0, I0=2, samplelen=10.0, restype='fast conv', coords='2θ', res=0.001,
                      incangle=0.0)

    # BEGIN Sample DO NOT CHANGE
    Amb = Layer(b=0, d=0.0, f=(1e-20 + 1e-20j), dens=1.0, magn_ang=0.0, sigma=0.0, xs_ai=0.0, magn=0.0)
    topPt = Layer(b=0, d=11.0, f=58, dens=4 / 3.92 ** 3, magn_ang=0.0, sigma=3.0, xs_ai=0.0, magn=0.0)
    TopFe = Layer(b=0, d=11.0, f=26, dens=2 / 2.866 ** 3, magn_ang=0.0, sigma=2.0, xs_ai=0.0, magn=0.0)
    Pt = Layer(b=0, d=11.0, f=58, dens=4 / 3.92 ** 3, magn_ang=0.0, sigma=2.0, xs_ai=0.0, magn=0.0)
    Fe = Layer(b=0, d=11, f=26, dens=2 / 2.866 ** 3, magn_ang=0.0, sigma=2.0, xs_ai=0.0, magn=0.0)
    bufPt = Layer(b=0, d=45, f=58, dens=4 / 3.92 ** 3, magn_ang=0.0, sigma=2, xs_ai=0.0, magn=0.0)
    bufFe = Layer(b=0, d=2, f=26, dens=2 / 2.866 ** 3, magn_ang=0.0, sigma=2, xs_ai=0.0, magn=0.0)
    Sub = Layer(b=0, d=0.0, f=12 + 16, dens=2 / 4.2 ** 3, magn_ang=0.0, sigma=4.0, xs_ai=0.0, magn=0.0)

    top = Stack(Layers=[TopFe, topPt], Repetitions=1)
    ML = Stack(Layers=[Fe, Pt], Repetitions=19)
    buffer = Stack(Layers=[bufFe, bufPt], Repetitions=1)

    sample = Sample(Stacks=[buffer, ML, top], Ambient=Amb, Substrate=Sub)

    # Test case for ReflFunction class
    from scipy import interpolate

    def create_dispersion_func(name='Fe'):
        path = '../databases/f1f2_nist/'
        e, f1, f2 = np.loadtxt(path + '%s.nff' % name.lower(), skiprows=1, \
                               unpack=True)
        f1interp = interpolate.interp1d(e, f1, kind='linear')
        f2interp = interpolate.interp1d(e, f2, kind='linear')

        def f(energy):
            return f1interp(energy) - 1.0J * f2interp(energy)

        return f


    fFe = ReflFunction(create_dispersion_func('Fe'), (1000,), {}, id = 'f(E)')
    fCo = ReflFunction(create_dispersion_func('Co'), (1000,), {}, id = 'f(E)')
    iprint(fFe.validate())
    iprint('Cast to array tests:')
    ltest = [fFe, 3.0, 3.0]
    iprint('Test single value: ', cast_to_array(ltest, 1000))
    iprint('Test array: ', cast_to_array(ltest, np.arange(1000, 1010)))
    iprint('Cast to array finished.')



#!/usr/bin/env python
#
# -*- Mode: python -*-
#
# $Id: Simplex.py,v 1.2 2004/05/31 14:01:06 vivake Exp $
# 
# Copyright (c) 2002-2004 Vivake Gupta (vivakeATlab49.com).  All rights reserved.
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
#
# This software is maintained by Vivake (vivakeATlab49.com) and is available at:
#     http://shell.lab49.com/~vivake/python/Simplex.py
#
# 1.2  ( 5/2004) - Fixed a bug found by Noboru Yamamoto <noboru.yamamotoATkek.jp>
#                  which caused minimize() not to converge, and reach the maxiter
#                  limit under some conditions.


""" Simplex - a regression method for arbitrary nonlinear function minimization

Simplex minimizes an arbitrary nonlinear function of N variables by the
Nedler-Mead Simplex method as described in:

Nedler, J.A. and Mead, R. "A Simplex Method for Function Minimization." 
    Computer Journal 7 (1965): 308-313.

It makes no assumptions about the smoothness of the function being minimized.
It converges to a local minimum which may or may not be the global minimum
depending on the initial guess used as a starting point.
"""

import math
import copy
from genx.gui_logging import iprint

class Simplex:
    def __init__(self, testfunc, guess, increments, kR = -1, kE = 2, kC = 0.5):
        """Initializes the simplex.
        INPUTS
        ------
        testfunc      the function to minimize
        guess[]       an list containing initial guesses
        increments[]  an list containing increments, perturbation size
        kR            reflection constant
        kE            expansion constant
        kC            contraction constant
        """
        self.testfunc = testfunc
        self.guess = guess
        self.increments = increments
        self.kR = kR
        self.kE = kE
        self.kC = kC
        self.numvars = len(self.guess)
        self.simplex = []

        self.lowest = -1
        self.highest = -1
        self.secondhighest = -1

        self.errors = []
        self.currenterror = 0

        # Initialize vertices
        for vertex in range(0, self.numvars + 3): # Two extras to store centroid and reflected point
            self.simplex.append(copy.copy(self.guess))
        # Use initial increments
        for vertex in range(0, self.numvars + 1):
            for x in range(0, self.numvars):
                if x == (vertex - 1):
                    self.simplex[vertex][x] = self.guess[x] + self.increments[x]
            self.errors.append(0)
        self.calculate_errors_at_vertices()

    def minimize(self, epsilon = 0.0001, maxiters = 250, monitor = 0):
        """Walks to the simplex down to a local minima.
        INPUTS
        ------
        epsilon       convergence requirement
        maxiters      maximum number of iterations
        monitor       if non-zero, progress info is output to stdout  

        OUTPUTS
        -------
        an array containing the final values
        lowest value of the error function
        number of iterations taken to get here
        """
        iter = 0
        for iter in range(0, maxiters):
            # Identify highest, second highest, and lowest vertices
            self.highest = 0
            self.lowest = 0
            for vertex in range(0, self.numvars + 1):
                if self.errors[vertex] > self.errors[self.highest]:
                    self.highest = vertex
                if self.errors[vertex] < self.errors[self.lowest]:
                    self.lowest = vertex
            self.secondhighest = 0
            for vertex in range(0, self.numvars + 1):
                if vertex == self.highest:
                    continue
                if self.errors[vertex] > self.errors[self.secondhighest]:
                    self.secondhighest = vertex
            # Test for convergence
            S = 0.0
            S1 = 0.0
            for vertex in range(0, self.numvars + 1):
                S = S + self.errors[vertex]
            F2 = S / (self.numvars + 1)
            for vertex in range(0, self.numvars + 1):
                S1 = S1 + (self.errors[vertex] - F2)**2
            T = math.sqrt(S1 / self.numvars)
            
            # Optionally, print progress information
            if monitor:
                iprint('Iteration = %d   Best = %f   Worst = %f' % (iter,self.errors[self.lowest],self.errors[self.highest]))
                
            if T <= epsilon:   # We converged!  Break out of loop!
                break;
            else:                   # Didn't converge.  Keep crunching.
                # Calculate centroid of simplex, excluding highest vertex
                for x in range(0, self.numvars):
                    S = 0.0
                    for vertex in range(0, self.numvars + 1):
                        if vertex == self.highest:
                            continue
                        S = S + self.simplex[vertex][x]
                    self.simplex[self.numvars + 1][x] = S / self.numvars

                self.reflect_simplex()

                self.currenterror = self.testfunc(self.guess)

                if self.currenterror < self.errors[self.lowest]:
                    tmp = self.currenterror
                    self.expand_simplex()
                    self.currenterror = self.testfunc(self.guess)
                    if self.currenterror < tmp:
                        self.accept_expanded_point()
                    else:
                        self.currenterror = tmp
                        self.accept_reflected_point()

                elif self.currenterror <= self.errors[self.secondhighest]:
                    self.accept_reflected_point()

                elif self.currenterror <= self.errors[self.highest]:
                    self.accept_reflected_point()

                    self.contract_simplex()
                    self.currenterror = self.testfunc(self.guess)
                    if self.currenterror < self.errors[self.highest]:
                        self.accept_contracted_point()
                    else:
                        self.multiple_contract_simplex()

                elif self.currenterror >= self.errors[self.highest]:
                    self.contract_simplex()
                    self.currenterror = self.testfunc(self.guess)
                    if self.currenterror < self.errors[self.highest]:
                        self.accept_contracted_point()
                    else:
                        self.multiple_contract_simplex()

        # Either converged or reached the maximum number of iterations.
        # Return the lowest vertex and the currenterror.
        for x in range(0, self.numvars):
            self.guess[x] = self.simplex[self.lowest][x]
        self.currenterror = self.errors[self.lowest]
        return self.guess, self.currenterror, iter

    def contract_simplex(self):
        for x in range(0, self.numvars):
            self.guess[x] = self.kC * self.simplex[self.highest][x] + (1 - self.kC) * self.simplex[self.numvars + 1][x]
        return

    def expand_simplex(self):
        for x in range(0, self.numvars):
            self.guess[x] = self.kE * self.guess[x]                 + (1 - self.kE) * self.simplex[self.numvars + 1][x]
        return

    def reflect_simplex(self):
        for x in range(0, self.numvars):
            self.guess[x] = self.kR * self.simplex[self.highest][x] + (1 - self.kR) * self.simplex[self.numvars + 1][x]
            self.simplex[self.numvars + 2][x] = self.guess[x] # REMEMBER THE REFLECTED POINT
        return

    def multiple_contract_simplex(self):
        for vertex in range(0, self.numvars + 1):
            if vertex == self.lowest:
                continue
            for x in range(0, self.numvars):
                self.simplex[vertex][x] = 0.5 * (self.simplex[vertex][x] + self.simplex[self.lowest][x])
        self.calculate_errors_at_vertices()
        return

    def accept_contracted_point(self):
        self.errors[self.highest] = self.currenterror
        for x in range(0, self.numvars):
            self.simplex[self.highest][x] = self.guess[x]
        return

    def accept_expanded_point(self):
        self.errors[self.highest] = self.currenterror
        for x in range(0, self.numvars):
            self.simplex[self.highest][x] = self.guess[x]
        return

    def accept_reflected_point(self):
        self.errors[self.highest] = self.currenterror
        for x in range(0, self.numvars):
            self.simplex[self.highest][x] = self.simplex[self.numvars + 2][x]
        return

    def calculate_errors_at_vertices(self):
        for vertex in range(0, self.numvars + 1):
            if vertex == self.lowest:
                continue
            for x in range(0, self.numvars):
                self.guess[x] = self.simplex[vertex][x]
            self.currenterror = self.testfunc(self.guess)
            self.errors[vertex] = self.currenterror
        return

def objective_function(args):
    return abs(args[0] * args[0] * args[0] * 5 - args[1] * args[1] * 7 + math.sqrt(abs(args[0])) - 118)

def main():
    s = Simplex(objective_function, [1, 1, 1], [2, 4, 6])
    values, err, iter = s.minimize(maxiters = 60, monitor = 0)
    iprint('args = ', values)
    iprint('error = ', err)
    iprint('iterations = ', iter)

if __name__ == '__main__':
    main()



from numpy import *
from functools import reduce
from genx.gui_logging import iprint

# "Ordinary" implementaion of Parrats recursion formula
# theta-vector, lamda- can be a vector,n-1Dvector, d-1Dvector, sigma-1Dvector
def Refl(theta, lamda, n, d, sigma, return_int=True):
    d = d[1:-1]
    sigma = sigma[:-1]
    # Length of k-vector in vaccum
    k = 2*math.pi/lamda
    # Calculates the wavevector in each layer
    Qj = 2*n[-1]*k*sqrt(n[:,newaxis]**2/n[-1]**2-cos(theta*math.pi/180)**2)
    # Fresnel reflectivity for the interfaces
    rp=(Qj[1:]-Qj[:-1])/(Qj[1:]+Qj[:-1])*exp(-Qj[1:]*Qj[:-1]/2*sigma[:,newaxis]**2)
    # Ignoring the top and bottom layer for the calc.
    p = exp(1.0j*d[:,newaxis]*Qj[1:-1])
    # Setting up a matrix for the reduce function. Reduce only takes one array
    # as argument
    rpp=array(list(map(lambda x,y:[x,y],rp[1:],p)))
    # Paratt's recursion formula
    def formula(rtot,rint):
        return (rint[0]+rtot*rint[1])/(1+rtot*rint[0]*rint[1])
    # Implement the recursion formula
    r = reduce(formula,rpp,rp[0])

    if return_int:
        return abs(r)**2
    else:
        return r

# "Ordinary" implementaion of Parrats recursion formula
# Q-vector,n-1Dvector, d-1Dvector, sigma-1Dvector
def ReflQ(Q,lamda,n,d,sigma, return_int=True):
    # Length of k-vector in vaccum
    d=d[1:-1]
    sigma=sigma[:-1]
    Q0=4*pi/lamda
    Q = Q.astype(complex128)
    # Calculates the wavevector in each layer
    Qj=sqrt((n[:,newaxis]**2 - n[-1]**2)*Q0**2 + n[-1]**2*Q**2)
    # Fresnel reflectivity for the interfaces
    rp=(Qj[1:]-Qj[:-1])/(Qj[1:]+Qj[:-1])*exp(-Qj[1:]*Qj[:-1]/2*sigma[:,newaxis]**2)
    #print rp.shape #For debugging
    #print d.shape
    #print Qj[1:-1].shape
    p=exp(1.0j*d[:,newaxis]*Qj[1:-1]) # Ignoring the top and bottom layer for the calc.
    #print p.shape #For debugging
    # Setting up a matrix for the reduce function. Reduce only takes one array
    # as argument
    rpp=array(list(map(lambda x,y:[x,y],rp[1:],p)))
    #print rpp.shape
    # Paratt's recursion formula
    def formula(rtot,rint):
        return (rint[0]+rtot*rint[1])/(1+rtot*rint[0]*rint[1])
    # Implement the recursion formula
    r=reduce(formula,rpp,rp[0])
    #print r.shape
    # return the reflectivity 
    if return_int:
        return abs(r)**2
    else:
        return r

# Parrats recursion formula for varying n given by n_func-function
def Refl_nvary(theta,lamda,n_func,d,sigma, return_int=True):
    d=d[1:-1]
    sigma=sigma[:-1]
    # Length of k-vector in vaccum
    k=2*math.pi/lamda
    # Calculates the wavevector in each layer
    #print n_func
    ss=transpose((sin(theta[:,newaxis]*pi/180.0)/lamda)*ones(len(n_func)))
    #print ss.shape
    #print theta.shape
    #print len(n_func)
    n=array(list(map(lambda f,val:f(val),n_func,ss)))
    #print n
    Qj=2*k*n[-1]*sqrt(n**2/n[-1]**2-cos(theta*math.pi/180)**2)
    # Fresnel reflectivity for the interfaces
    rp=(Qj[1:]-Qj[:-1])/(Qj[1:]+Qj[:-1])*exp(-Qj[1:]*Qj[:-1]/2*sigma[:,newaxis]**2)
    #print rp.shape #For debugging
    #print d.shape
    #print Qj[1:-1].shape
    p=exp(1.0j*d[:,newaxis]*Qj[1:-1]) # Ignoring the top and bottom layer for the calc.
    #print p.shape #For debugging
    # Setting up a matrix for the reduce function. Reduce only takes one array
    # as argument
    rpp=array(list(map(lambda x,y:[x,y],rp[1:],p)))
    #print rpp.shape
    # Paratt's recursion formula
    def formula(rtot,rint):
        return (rint[0]+rtot*rint[1])/(1+rtot*rint[0]*rint[1])
    # Implement the recursion formula
    r=reduce(formula,rpp,rp[0])
    if return_int:
        return abs(r)**2
    else:
        return r

def ReflProfiles(theta,lamda,n,d,sigma,profile, return_int=True):
    d=d[1:-1]
    sigma=sigma[:-1]
    # Length of k-vector in vaccum
    k=2*math.pi/lamda
    # Calculates the wavevector in each layer
    Qj=2*n[-1]*k*sqrt(n[:,newaxis]**2/n[-1]**2 - cos(theta*math.pi/180)**2)
    # Function to map caclulate the roughness values
    def w(QiQj,sigma,profile):
        # erf profile
        if profile==0:
            return exp(-QiQj*sigma**2/2)
        # exponential profile
        elif profile==1:
            return 1/(1+sqrt(QiQj)*sigma**2/2)
        # linear profile
        elif profile==2:
            return sin(sqrt(3)*sigma*sqrt(QiQj))/sqrt(3)/sigma/sqrt(QiQj)
        else:
            # If any wrong number are typed in return the erf profile
            return exp(-QiQj*sigma**2/2)

    interface=array(list(map(w,Qj[:-1]*Qj[1:],sigma,profile)))
    # Fresnel reflectivity for the interfaces
    rp=(Qj[1:]-Qj[:-1])/(Qj[1:]+Qj[:-1])*interface
    #print rp.shape #For debugging
    #print d.shape
    #print Qj[1:-1].shape
    p=exp(1.0j*d[:,newaxis]*Qj[1:-1]) # Ignoring the top and bottom layer for the calc.
    #print p.shape #For debugging
    # Setting up a matrix for the reduce function. Reduce only takes one array
    # as argument
    rpp=array(list(map(lambda x,y:[x,y],rp[1:],p)))
    #print rpp.shape
    # Paratt's recursion formula
    def formula(rtot,rint):
        return (rint[0]+rtot*rint[1])/(1+rtot*rint[0]*rint[1])
    # Implement the recursion formula
    r=reduce(formula,rpp,rp[0])
    #print r.shape
    # return the reflectivity 
    if return_int:
        return abs(r)**2
    else:
        return r

# paratts algorithm for n as function of lamda or theta
def Refl_nvary2(theta,lamda,n_vector,d,sigma, return_int=True):
    d=d[1:-1]
    sigma=sigma[:-1]
    # Length of k-vector in vaccum
    k=2*math.pi/lamda
    # Calculates the wavevector in each layer
    #print n_func
    #ss=transpose((sin(theta[:,newaxis]*pi/180.0)/lamda)*ones(len(n_func)))
    #print ss.shape
    #print theta.shape
    #print len(n_func)
    #n=array(map(lambda f,val:f(val),n_func,ss))
    n=n_vector
    #print n
    Qj=2*n[-1]*k*sqrt(n**2/n[-1]**2-cos(theta*math.pi/180)**2)
    #print sigma.shape, Qj.shape
    # Fresnel reflectivity for the interfaces
    rp=(Qj[1:]-Qj[:-1])/(Qj[1:]+Qj[:-1])*exp(-Qj[1:]*Qj[:-1]/2*sigma[:,newaxis]**2)
    #print rp.shape #For debugging
    #print d.shape
    #print Qj[1:-1].shape
    p=exp(1.0j*d[:,newaxis]*Qj[1:-1]) # Ignoring the top and bottom layer for the calc.
    #print p.shape #For debugging
    # Setting up a matrix for the reduce function. Reduce only takes one array
    # as argument
    rpp=array(list(map(lambda x,y:[x,y],rp[1:],p)))
    #print rpp.shape
    # Paratt's recursion formula
    def formula(rtot,rint):
        return (rint[0]+rtot*rint[1])/(1+rtot*rint[0]*rint[1])
    # Implement the recursion formula
    r=reduce(formula,rpp,rp[0])
    if return_int:
        return abs(r)**2
    else:
        return r


def reflq_kin(q, lamda, n, d, sigma, correct_q=True, return_int=True):
    """Calculates the reflectivity in the kinematical approximation"""
    d = d[:-1]
    d[0] = 0
    z = d.sum() - d.cumsum()
    sigma = sigma[:-1]
    q0 = 4*pi/lamda
    # Kinematical reflectivity for the interfaces
    iprint(n.shape, len(n.shape))
    if len(n.shape) == 1:
        if correct_q:
            # The internal wave vector calacuted with the thickness averaged refractive index.
            n_mean = (n[1:-1] * d[1:] / d.sum()).sum()
            q_corr = sqrt((n_mean ** 2 - n[-1] ** 2) * q0 ** 2 + (n[-1] * q) ** 2)
        else:
            q_corr = q
        rp = (n[:-1] - n[1:])[:, newaxis]*exp(-(q_corr*sigma[:, newaxis])**2/2)
    else:
        if correct_q:
            # The internal wave vector calacuted with the thickness averaged refractive index.
            n_mean = (n[1:-1] * d[1:][:, newaxis] / d.sum()).sum(axis=0)
            q_corr = sqrt((n_mean ** 2 - n[-1] ** 2) * q0 ** 2 + (n[-1] * q) ** 2)
        else:
            q_corr = q
        rp = (n[:-1] - n[1:]) * exp(-(q_corr * sigma[:, newaxis]) ** 2 / 2)
    p = exp(1.0j*z[:, newaxis]*q_corr)

    r = (rp*p).sum(axis=0)*q0**2/q_corr**2/2.

    if return_int:
        return abs(r)**2
    else:
        return r


def reflq_pseudo_kin(q, lamda, n, d, sigma, return_int=True):
    """Calculates the reflectivity in a pseudo kinematical approximation.
    The mean refractive index of the film is simulated with the single reflection approximation and the deviation from
    the mean is simulated with the kinematical approximation.
    """
    d = d[:-1]
    d[0] = 0
    z = d.sum() - d.cumsum()
    sigma = sigma[:-1]
    q0 = 4*pi/lamda
    # Q = Q.astype(complex128)
    # The internal wave vector calacuted with the thickness averaged refractive index.
    n_mean = (n[1:-1]*d[1:]/d.sum()).sum()
    q_corr = sqrt((n_mean**2 - n[-1]**2)*q0**2 + n[-1]**2*q**2)
    q_sub = sqrt((n[0]**2 - n[-1]**2)*q0**2 + n[-1]**2*q**2)
    q_amb = n[-1]*q
    # Top interface
    rp_top = (q_corr - q_amb) / (q_corr + q_amb) * exp(-q_corr*q/2 * sigma[-1]**2)
    rp_sub = (q_sub - q_corr) / (q_sub + q_corr) * exp(-q_sub*q_corr/2 * sigma[0]**2)
    #rp_top = -(n[-1] - n_mean)*exp(-(q_corr*sigma[-1])**2/2)*q0**2/q_corr**2/2.
    #rp_sub = -(n_mean - n[0])*exp(-(q_corr*sigma[0])**2/2)*q0**2/q_corr**2/2.
    # Kinematical reflectivity for the interfaces
    n_diff = n - n_mean
    n_diff[0] = 0
    n_diff[-1] = 0
    rp = (n_diff[:-1] - n_diff[1:])[:, newaxis]*exp(-(q_corr*sigma[:, newaxis])**2/2)
    p = exp(1.0j*z[:, newaxis]*q_corr)

    r_kin = (rp*p).sum(axis=0)*q0**2/q_corr**2/2.
    r_sra = rp_top + rp_sub*exp(1.0j*d.sum()*q_corr)
    r = r_kin + r_sra

    if return_int:
        return abs(r)**2
    else:
        return r


def reflq_sra(q, lamda, n, d, sigma, return_int=True):
    """Single reflection approximation calculation of the reflectivity"""
    # Length of k-vector in vaccum
    d = d[1:-1]
    sigma=sigma[:-1]
    q0 = 4*pi/lamda
    # Calculates the wavevector in each layer
    if len(n.shape) == 1:
        qj = sqrt((n[:, newaxis]**2 - n[-1]**2)*q0**2 + (n[-1]*q)**2)
    else:
        qj = sqrt((n ** 2 - n[-1] ** 2) * q0 ** 2 + (n[-1] * q) ** 2)
    # Fresnel reflectivity for the interfaces
    rp = (qj[:-1] - qj[1:])/(qj[1:] + qj[:-1])*exp(-qj[1:]*qj[:-1]/2*sigma[:, newaxis]**2)
    # The wave does not transverse the ambient and substrate - ignoring them
    # Also, the wave travels from top -> bottom, the array has the first element as the substrate
    # - need to reverse the order.
    phaseterm = (d[:, newaxis]*qj[1:-1])[::-1].cumsum(axis=0)[::-1]
    p = exp(1.0J*phaseterm)
    # Adding the first interface (top -> last in array) since p is not calculated for that layer (p = 1)
    r = rp[-1] + (rp[:-1]*p).sum(axis=0)
    if return_int:
        return abs(r)**2
    else:
        return r

from . import USE_NUMBA
if USE_NUMBA:
    # try to use numba to speed up the calculation intensive functions:
    try:
        from .paratt_numba import Refl, ReflQ, Refl_nvary2
    except Exception as e:
        iprint('Could not use numba, no speed up from JIT compiler:\n'+str(e))

if __name__=='__main__':
    theta=arange(0,5,0.01)+1e-13
    #c=paratt.Refl(theta,1.54,array([1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1]),array([80,20,80,20,80,20]),0)
    #c1=Refl(theta,1.54,array([1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1]),array([80,20,80,20,80,20]),4)
    c2=Refl(theta,1.54,array([1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1]),array([2,80,20,80,20,80,20,2]),array([0,0,0,0,0,0,0,0]))
    #print c.shape
    #gplt.plot(theta,log10(c1),theta,log10(c2))
    #gplt.plot(theta,log10(c2))



import numpy as np
from genx.gui_logging import iprint

_ctype = np.complex128

def refl(kj, kjm1):
    return (kj + kjm1)/2/kjm1

def ass_X_test(k_b, k_u, k_l, dd_u, dd_l,
                        sigma, sigma_l, sigma_u, dtype = _ctype):
    ''' Just a simpler version of ass_X to locate an error somewhere...
    '''
    k_jm1_b = k_b[...,1:]
    k_j_b = k_b[...,:-1]
    k_jm1_l = k_l[...,1:]
    k_j_u = k_u[...,:-1]
    sigma_jm1_l = sigma_l[...,1:]
    sigma_j_b = sigma[...,:-1]
    sigma_j_u = sigma_u[...,:-1]
    dd_jm1_l = dd_l[...,1:]
    dd_j_u = dd_u[...,:-1]
    
    iprint(dd_j_u, dd_jm1_l)

    #Reflectivites
    r_j_b = refl(k_j_b, k_j_u)
    #print 'r_j_b', r_j_b
    r_jm1_l = refl(k_jm1_l, k_jm1_b) 
    #print 'r_jm1_l', r_jm1_l
    r_j_u = refl(k_j_u, k_jm1_l)
    #print 'r_j_u', r_j_u
    p_jm1_l = np.exp(-1.0J*dd_jm1_l*k_jm1_l)
    p_j_u = np.exp(-1.0J*dd_j_u*k_j_u)
    
    #Defining the X matrix
    X = np.empty((2,2) + k_j_b.shape, dtype = dtype)
    
    X[0,0] = p_jm1_l*r_j_u*p_j_u
    X[0,1] = -(p_jm1_l*r_j_u-p_jm1_l)/p_j_u
    X[1,0] = -((r_j_u-1)*p_j_u)/p_jm1_l
    X[1,1] = r_j_u/(p_jm1_l*p_j_u)
    
    return X


def ass_X_interfacelayer4(k_b, k_u, k_l, dd_u, dd_l,
                        sigma, sigma_l, sigma_u, dtype = _ctype):
    '''Function that creates a interfacelayer at the top and bottom
    of the layer. with correlated roughness for the interface layers and the
    additional roughness of sigma_l and sigma_u.
    The subscripts b, l and u reffers to bulk, lower and upper respectively.

    v 0.4
    '''
    #print dd_l, dd_u
    # Variables definitions that goes into the expressions:
    k_jm1_b = k_b[...,1:]
    k_j_b = k_b[...,:-1]
    k_jm1_l = k_l[...,1:]
    k_j_u = k_u[...,:-1]
    sigma_jm1_l = sigma_l[...,1:]
    sigma_j_b = sigma[...,:-1]
    sigma_j_u = sigma_u[...,:-1]
    dd_jm1_l = dd_l[...,1:]
    dd_j_u = dd_u[...,:-1]

    #Reflectivites
    r_j_b = refl(k_j_b, k_j_u)
    r_jm1_l = refl(k_jm1_l, k_jm1_b) 
    r_j = refl(k_j_u, k_jm1_l)
    p_jm1_l = np.exp(-1.0J*dd_jm1_l*k_jm1_l)
    p_j_u = np.exp(-1.0J*dd_j_u*k_j_u)
    #p_j_b = np.exp(-1.0J*dd_j_b*k_j_b)

    #Defining the X matrix
    X = np.empty((2,2) + k_j_b.shape, dtype = dtype)

    tmp1=k_jm1_b**2
    tmp2=sigma_jm1_l**2
    tmp3=-(tmp1*tmp2)/2
    tmp4=-k_jm1_b*tmp2*k_jm1_l
    tmp5=-(tmp2*k_jm1_l**2)/2
    tmp6=sigma_j_b**2
    tmp7=-(tmp1*tmp6)/2
    tmp8=k_j_b**2
    tmp9=-(tmp6*tmp8)/2
    tmp10=sigma_j_u**2
    tmp11=-(tmp8*tmp10)/2
    tmp12=-k_j_b*tmp10*k_j_u
    tmp13=-(tmp10*k_j_u**2)/2
    tmp14=np.exp(tmp13+tmp12+tmp11+tmp9+tmp7+tmp5+tmp4+tmp3)
    tmp15=1/p_jm1_l
    tmp16=1/p_j_u
    tmp17=2*k_jm1_b*tmp2*k_jm1_l
    tmp18=k_jm1_b*tmp6*k_j_b
    tmp21=np.exp(tmp18)
    tmp19=np.exp(tmp17)*tmp21
    tmp20=p_jm1_l**2
    #tmp21=np.exp(tmp18)
    tmp22=-tmp21
    tmp23=2*k_j_b*tmp10*k_j_u
    #tmp24=np.exp(tmp23+tmp18)
    #tmp24=np.exp(tmp23)*tmp21
    #tmp25=np.exp(tmp23+tmp18+tmp17)
    tmp26=p_j_u**2
    #tmp27=np.exp(tmp13+tmp12+tmp11+tmp9-k_jm1_b*tmp6*k_j_b+tmp7+tmp5+tmp4+tmp3)
    tmp27 = tmp14/tmp21
    tmp28=np.exp(tmp17)
    #tmp29=np.exp(tmp23+tmp17)
    tmp30=np.exp(tmp23)
    tmp29 = tmp28*tmp30
    tmp24=tmp30*tmp21
    tmp25=tmp30*tmp19
    tmp31=-tmp20
    tmp32=tmp21*tmp20
    tmp33=-tmp21*tmp20
    tmp34=tmp24*tmp20

    X[0,0] = tmp14*tmp15*tmp16*(((((tmp25*tmp20+tmp24)*r_jm1_l-tmp24)*r_j-tmp24*r_jm1_l+tmp24)*tmp26+((tmp19*tmp20+tmp21)*r_jm1_l+tmp22)*r_j-tmp19*tmp20*r_jm1_l)*r_j_b+((tmp22-tmp19*tmp20)*r_jm1_l+tmp21)*r_j+tmp19*tmp20*r_jm1_l)
    
    X[0,1] = -tmp27*tmp15*tmp16*(((((tmp28*tmp20+1)*r_jm1_l-1)*r_j-r_jm1_l+1)*tmp26+((tmp29*tmp20+tmp30)*r_jm1_l-tmp30)*r_j-tmp29*tmp20*r_jm1_l)*r_j_b+(((-tmp28*tmp20-1)*r_jm1_l+1)*r_j+r_jm1_l-1)*tmp26)
    
    X[1,0] = -tmp27*tmp15*tmp16*(((((tmp30*tmp20+tmp29)*r_jm1_l-tmp30*tmp20)*r_j-tmp29*r_jm1_l)*tmp26+((tmp20+tmp28)*r_jm1_l+tmp31)*r_j-tmp20*r_jm1_l+tmp20)*r_j_b+((tmp31-tmp28)*r_jm1_l+tmp20)*r_j+tmp20*r_jm1_l+tmp31)
    
    X[1,1] = tmp14*tmp15*tmp16*(((((tmp32+tmp19)*r_jm1_l+tmp33)*r_j-tmp19*r_jm1_l)*tmp26+((tmp34+tmp25)*r_jm1_l-tmp24*tmp20)*r_j-tmp24*tmp20*r_jm1_l+tmp34)*r_j_b+(((tmp33-tmp19)*r_jm1_l+tmp32)*r_j+tmp19*r_jm1_l)*tmp26)

    
    return X

def ass_X_interfacelayer3(k_b, k_u, k_l, dd_u, dd_l,
                        sigma, sigma_l, sigma_u, dtype = _ctype):
    '''Function that creates a interfacelayer at the top and bottom
    of the layer. with correlated roughness for the interface layers and the
    additional roughness of sigma_l and sigma_u.
    The subscripts b, l and u reffers to bulk, lower and upper respectively.

    v 0.3
    '''

    # Variables definitions that goes into the expressions:
    k_jm1_b = k_b[...,1:]
    k_j_b = k_b[...,:-1]
    k_jm1_l = k_l[...,1:]
    k_j_u = k_u[...,:-1]
    sigma_jm1_l = sigma_l[...,1:]
    sigma_j_b = sigma[...,:-1]
    sigma_j_u = sigma_u[...,:-1]
    dd_jm1_l = dd_l[...,1:]
    dd_j_u = dd_u[...,:-1]
    
    iprint(dd_j_u, dd_jm1_l)

    #Reflectivites
    r_j_b = refl(k_j_b, k_j_u)
    iprint('r_j_b', r_j_b)
    r_jm1_l = refl(k_jm1_l, k_jm1_b) 
    iprint('r_jm1_l', r_jm1_l)
    r_j_u = refl(k_j_u, k_jm1_l)
    iprint('r_j_u', r_j_u)
    p_jm1_l = np.exp(-1.0J*dd_jm1_l*k_jm1_l)
    p_j_u = np.exp(-1.0J*dd_j_u*k_j_u)
    #p_j_b = np.exp(-1.0J*dd_j_b*k_j_b)

    #Defining the X matrix
    X = np.empty((2,2) + k_j_b.shape, dtype = dtype)

    tmp1=k_jm1_b**2
    tmp2=sigma_jm1_l**2
    tmp3=-(tmp1*tmp2)/2
    tmp4=k_jm1_b*tmp2*k_jm1_l
    tmp5=-(tmp2*k_jm1_l**2)/2
    tmp6=sigma_j_b**2
    tmp7=-(tmp1*tmp6)/2
    tmp8=k_jm1_b*tmp6*k_j_b
    tmp9=k_j_b**2
    tmp10=-(tmp6*tmp9)/2
    tmp11=sigma_j_u**2
    tmp12=-(tmp9*tmp11)/2
    tmp13=-k_j_b*tmp11*k_j_u
    tmp14=-(tmp11*k_j_u**2)/2
    tmp15=np.exp(tmp14+tmp13+tmp12+tmp10+tmp8+tmp7+tmp5+tmp4+tmp3)
    tmp16=1/p_j_u
    tmp17=-k_jm1_b*tmp2*k_jm1_l
    tmp18=np.exp(tmp14+tmp13+tmp12+tmp10+tmp8+tmp7+tmp5+tmp17+tmp3)
    tmp19=1/p_jm1_l
    tmp20=k_j_b*tmp11*k_j_u
    tmp21=np.exp(tmp14+tmp20+tmp12+tmp10+tmp8+tmp7+tmp5+tmp17+tmp3)
    tmp22=np.exp(tmp14+tmp20+tmp12+tmp10+tmp8+tmp7+tmp5+tmp4+tmp3)
    tmp23=-k_jm1_b*tmp6*k_j_b
    tmp24=np.exp(tmp14+tmp13+tmp12+tmp10+tmp23+tmp7+tmp5+tmp17+tmp3)
    tmp25=np.exp(tmp14+tmp13+tmp12+tmp10+tmp23+tmp7+tmp5+tmp4+tmp3)
    tmp26=np.exp(tmp14+tmp20+tmp12+tmp10+tmp23+tmp7+tmp5+tmp4+tmp3)
    tmp27=np.exp(tmp14+tmp20+tmp12+tmp10+tmp23+tmp7+tmp5+tmp17+tmp3)

     
    X[0,0] = tmp22*p_jm1_l*r_jm1_l*r_j_b*p_j_u*r_j_u+tmp21*tmp19*r_jm1_l*r_j_b*p_j_u*r_j_u-tmp21*tmp19*r_j_b*p_j_u*r_j_u-tmp21*tmp19*r_jm1_l*p_j_u*r_j_u+tmp21*tmp19*p_j_u*r_j_u+tmp15*p_jm1_l*r_jm1_l*r_j_b*tmp16*r_j_u+tmp18*tmp19*r_jm1_l*r_j_b*tmp16*r_j_u-tmp18*tmp19*r_j_b*tmp16*r_j_u-tmp15*p_jm1_l*r_jm1_l*tmp16*r_j_u-tmp15*p_jm1_l*r_jm1_l*r_j_b*tmp16-tmp18*tmp19*r_jm1_l*r_j_b*tmp16+tmp18*tmp19*r_j_b*tmp16+tmp15*p_jm1_l*r_jm1_l*tmp16
    
    X[0,1] = -tmp25*p_jm1_l*r_jm1_l*r_j_b*p_j_u*r_j_u-tmp24*tmp19*r_jm1_l*r_j_b*p_j_u*r_j_u+tmp24*tmp19*r_j_b*p_j_u*r_j_u+tmp24*tmp19*r_jm1_l*p_j_u*r_j_u-tmp24*tmp19*p_j_u*r_j_u-tmp26*p_jm1_l*r_jm1_l*r_j_b*tmp16*r_j_u-tmp27*tmp19*r_jm1_l*r_j_b*tmp16*r_j_u+tmp27*tmp19*r_j_b*tmp16*r_j_u+tmp26*p_jm1_l*r_jm1_l*tmp16*r_j_u+tmp25*p_jm1_l*r_jm1_l*r_j_b*p_j_u+tmp24*tmp19*r_jm1_l*r_j_b*p_j_u-tmp24*tmp19*r_j_b*p_j_u-tmp24*tmp19*r_jm1_l*p_j_u+tmp24*tmp19*p_j_u
    
    X[1,0] = -tmp27*p_jm1_l*r_jm1_l*r_j_b*p_j_u*r_j_u-tmp26*tmp19*r_jm1_l*r_j_b*p_j_u*r_j_u+tmp27*p_jm1_l*r_j_b*p_j_u*r_j_u+tmp26*tmp19*r_jm1_l*p_j_u*r_j_u-tmp24*p_jm1_l*r_jm1_l*r_j_b*tmp16*r_j_u-tmp25*tmp19*r_jm1_l*r_j_b*tmp16*r_j_u+tmp24*p_jm1_l*r_j_b*tmp16*r_j_u+tmp24*p_jm1_l*r_jm1_l*tmp16*r_j_u-tmp24*p_jm1_l*tmp16*r_j_u+tmp24*p_jm1_l*r_jm1_l*r_j_b*tmp16+tmp25*tmp19*r_jm1_l*r_j_b*tmp16-tmp24*p_jm1_l*r_j_b*tmp16-tmp24*p_jm1_l*r_jm1_l*tmp16+tmp24*p_jm1_l*tmp16
    
    X[1,1] = tmp18*p_jm1_l*r_jm1_l*r_j_b*p_j_u*r_j_u+tmp15*tmp19*r_jm1_l*r_j_b*p_j_u*r_j_u-tmp18*p_jm1_l*r_j_b*p_j_u*r_j_u-tmp15*tmp19*r_jm1_l*p_j_u*r_j_u+tmp21*p_jm1_l*r_jm1_l*r_j_b*tmp16*r_j_u+tmp22*tmp19*r_jm1_l*r_j_b*tmp16*r_j_u-tmp21*p_jm1_l*r_j_b*tmp16*r_j_u-tmp21*p_jm1_l*r_jm1_l*tmp16*r_j_u+tmp21*p_jm1_l*tmp16*r_j_u-tmp18*p_jm1_l*r_jm1_l*r_j_b*p_j_u-tmp15*tmp19*r_jm1_l*r_j_b*p_j_u+tmp18*p_jm1_l*r_j_b*p_j_u+tmp15*tmp19*r_jm1_l*p_j_u
    
    
    
    return X



def ass_X_interfacelayer(k_b, k_u, k_l, dd_u, dd_l,
                        sigma, sigma_l, sigma_u, dtype = _ctype):
    '''Function that creates a interfacelayer at the top and bottom
    of the layer. with correlated roughness for the interface layers and the
    additional roughness of sigma_l and sigma_u.
    The subscripts b, l and u reffers to bulk, lower and upper respectively.

    v 0.2
    '''
    iprint(dd_l, dd_u)
    # Variables definitions that goes into the expressions:
    k_jm1_b = k_b[...,1:]
    k_j_b = k_b[...,:-1]
    k_jm1_l = k_l[...,1:]
    k_j_u = k_u[...,:-1]
    sigma_jm1_l = sigma_l[...,1:]
    sigma_j_b = sigma[...,:-1]
    sigma_j_u = sigma_u[...,:-1]
    dd_jm1_l = dd_l[...,1:]
    dd_j_u = dd_u[...,:-1]

    #Reflectivites
    r_j_b = refl(k_j_b, k_j_u)
    r_jm1_l = refl(k_jm1_l, k_jm1_b) 
    r_j_u = refl(k_j_u, k_jm1_l)
    p_jm1_l = np.exp(-1.0J*dd_jm1_l*k_jm1_b)
    p_j_u = np.exp(-1.0J*dd_j_u*k_j_u)
    #p_j_b = np.exp(-1.0J*dd_j_b*k_j_b)

    #Defining the X matrix
    X = np.empty((2,2) + k_j_b.shape, dtype = dtype)

    tmp1=k_jm1_b**2
    tmp2=sigma_jm1_l**2
    tmp3=-(tmp1*tmp2)/2
    tmp4=-k_jm1_b*tmp2*k_jm1_l
    tmp5=-(tmp2*k_jm1_l**2)/2
    tmp6=sigma_j_b**2
    tmp7=-(tmp1*tmp6)/2
    tmp8=k_j_b**2
    tmp9=-(tmp6*tmp8)/2
    tmp10=sigma_j_u**2
    tmp11=-(tmp8*tmp10)/2
    tmp12=-k_j_b*tmp10*k_j_u
    tmp13=-(tmp10*k_j_u**2)/2
    tmp14=np.exp(tmp13+tmp12+tmp11+tmp9+tmp7+tmp5+tmp4+tmp3)
    tmp15=1/p_jm1_l
    tmp16=1/p_j_u
    tmp17=2*k_jm1_b*tmp2*k_jm1_l
    tmp18=k_jm1_b*tmp6*k_j_b
    tmp21=np.exp(tmp18)
    tmp19=np.exp(tmp17)*tmp21
    tmp20=p_jm1_l**2
    #tmp21=np.exp(tmp18)
    tmp22=-tmp21
    tmp23=2*k_j_b*tmp10*k_j_u
    #tmp24=np.exp(tmp23+tmp18)
    #tmp24=np.exp(tmp23)*tmp21
    #tmp25=np.exp(tmp23+tmp18+tmp17)
    tmp26=p_j_u**2
    #tmp27=np.exp(tmp13+tmp12+tmp11+tmp9-k_jm1_b*tmp6*k_j_b+tmp7+tmp5+tmp4+tmp3)
    tmp27 = tmp14/tmp21
    tmp28=np.exp(tmp17)
    #tmp29=np.exp(tmp23+tmp17)
    tmp30=np.exp(tmp23)
    tmp29 = tmp28*tmp30
    tmp24=tmp30*tmp21
    tmp25=tmp30*tmp19
    tmp31=-tmp20
    tmp32=tmp21*tmp20
    tmp33=-tmp21*tmp20
    tmp34=tmp24*tmp20

    X[0,0] = tmp14*tmp15*tmp16*(((((tmp25*tmp20+tmp24)*r_jm1_l-tmp24)*r_j_b-tmp24*r_jm1_l+tmp24)*tmp26+((tmp19*tmp20+tmp21)*r_jm1_l+tmp22)*r_j_b-tmp19*tmp20*r_jm1_l)*r_j_u+((tmp22-tmp19*tmp20)*r_jm1_l+tmp21)*r_j_b+tmp19*tmp20*r_jm1_l)
    
    X[0,1] = -tmp27*tmp15*tmp16*(((((tmp28*tmp20+1)*r_jm1_l-1)*r_j_b-r_jm1_l+1)*tmp26+((tmp29*tmp20+tmp30)*r_jm1_l-tmp30)*r_j_b-tmp29*tmp20*r_jm1_l)*r_j_u+(((-tmp28*tmp20-1)*r_jm1_l+1)*r_j_b+r_jm1_l-1)*tmp26)
    
    X[1,0] = -tmp27*tmp15*tmp16*(((((tmp30*tmp20+tmp29)*r_jm1_l-tmp30*tmp20)*r_j_b-tmp29*r_jm1_l)*tmp26+((tmp20+tmp28)*r_jm1_l+tmp31)*r_j_b-tmp20*r_jm1_l+tmp20)*r_j_u+((tmp31-tmp28)*r_jm1_l+tmp20)*r_j_b+tmp20*r_jm1_l+tmp31)
    
    X[1,1] = tmp14*tmp15*tmp16*(((((tmp32+tmp19)*r_jm1_l+tmp33)*r_j_b-tmp19*r_jm1_l)*tmp26+((tmp34+tmp25)*r_jm1_l-tmp24*tmp20)*r_j_b-tmp24*tmp20*r_jm1_l+tmp34)*r_j_u+(((tmp33-tmp19)*r_jm1_l+tmp32)*r_j_b+tmp19*r_jm1_l)*tmp26)

    return X

def ass_X_interfacelayer1(k_b, k_u, k_l, dd_u, dd_l,
                        sigma, sigma_l, sigma_u, dtype = _ctype):
    '''Function that creates a interfacelayer at the top and bottom
    of the layer. with correlated roughness for the interface layers and the
    additional roughness of sigma_l and sigma_u.
    The subscripts b, l and u reffers to bulk, lower and upper respectively.

    v 0.1
    '''

    # Variables definitions that goes into the expressions:
    k_jm1_b = k_b[...,1:]
    k_j_b = k_b[...,:-1]
    k_jm1_l = k_l[...,1:]
    k_j_u = k_u[...,:-1]
    sigma_jm1_l = sigma_l[...,1:]
    sigma_j_b = sigma[...,:-1]
    sigma_j_u = sigma_u[...,:-1]
    dd_jm1_l = dd_u[...,1:]
    dd_j_u = dd_u[...,:-1]

    #Defining the X matrix
    X = np.empty((2,2) + k_j_b.shape, dtype = dtype)

    # Temporary variables to ease the calculation
    tmp1 = k_jm1_b*k_jm1_b
    tmp2 = sigma_jm1_l*sigma_jm1_l
    tmp3 = tmp1*tmp2
    tmp4 = (2*k_jm1_b*tmp2+2*1.0J*dd_jm1_l)*k_jm1_l
    tmp5 = k_jm1_l*k_jm1_l
    tmp6 = tmp2*tmp5
    tmp7 = sigma_j_b*sigma_j_b
    tmp8 = tmp1*tmp7
    tmp9 = k_j_b**2
    tmp10 = tmp7*tmp9
    tmp11 = sigma_j_u*sigma_j_u
    tmp12 = tmp9*tmp11
    tmp13 = (2*k_j_b*tmp11+2*1.0J*dd_j_u)*k_j_u
    tmp14 = k_j_u*k_j_u
    tmp15 = tmp11*tmp14
    tmp16 = 1/np.exp((tmp15+tmp13+tmp12+tmp10+tmp8+tmp6+tmp4+tmp3)/2)
    tmp17 = 1/k_jm1_b
    tmp18 = 1/k_jm1_l
    tmp19 = 1/k_j_u
    tmp20 = np.exp(2*1.0J*dd_jm1_l*k_jm1_l)
    tmp21 = -tmp20
    tmp22 = np.exp(2*k_jm1_b*tmp2*k_jm1_l)
    tmp23 = -tmp22
    tmp24 = tmp23+tmp21
    tmp25 = k_jm1_b*tmp7*k_j_b
    tmp26 = 2*1.0J*dd_j_u*k_j_u
    tmp27 = np.exp(tmp26+tmp25)
    tmp28 = tmp22+tmp20
    tmp29 = 2*k_j_b*tmp11*k_j_u
    tmp30 = np.exp(tmp29+tmp25)
    tmp31 = tmp28*tmp30
    tmp32 = tmp31+tmp24*tmp27
    tmp33 = tmp23+tmp20
    tmp34 = tmp22+tmp21
    tmp35 = tmp34*tmp30
    tmp36 = tmp35+tmp33*tmp27
    tmp37 = tmp31+tmp28*tmp27
    tmp38 = tmp35+tmp34*tmp27
    tmp39 = 1/np.exp((tmp15+tmp13+tmp12+tmp10+2*k_jm1_b*tmp7*k_j_b+tmp8+tmp6+tmp4+tmp3)/2)
    tmp40 = np.exp(tmp13)
    tmp41 = tmp24*tmp40
    tmp42 = tmp33*tmp40
    tmp43 = tmp34*tmp40
    tmp44 = tmp28*tmp40
    tmp45 = np.exp(tmp4)
    tmp46 = tmp45+1
    tmp47 = np.exp(tmp26)
    tmp48 = tmp46*tmp47
    tmp49 = -tmp45
    tmp50 = tmp49-1
    tmp51 = np.exp(tmp29)
    tmp52 = tmp50*tmp51
    tmp53 = tmp45-1
    tmp54 = tmp53*tmp47
    tmp55 = tmp49+1
    tmp56 = tmp55*tmp51
    tmp57 = tmp50*tmp47
    tmp58 = tmp55*tmp47
    tmp59 = tmp53*tmp51
    tmp60 = tmp46*tmp51
    tmp61 = np.exp(tmp25)
    tmp62 = np.exp(tmp13+tmp25)
    tmp63 = tmp46*tmp62
    tmp64 = tmp63+tmp50*tmp61
    tmp65 = tmp53*tmp62
    tmp66 = tmp65+tmp55*tmp61
    tmp67 = tmp63+tmp46*tmp61
    tmp68 = tmp65+tmp53*tmp61


    # Assembling the interface matrices
    X[0,0] = (tmp16*tmp17*tmp18*tmp19*((tmp32*k_jm1_l+tmp36*k_jm1_b)*tmp14+((tmp37*k_jm1_l+tmp38*k_jm1_b)*k_j_b+tmp38*tmp5+tmp37*k_jm1_b*k_jm1_l)*k_j_u+(tmp36*tmp5+tmp32*k_jm1_b*k_jm1_l)*k_j_b))/8

    X[0,1] = -(tmp39*tmp17*tmp18*tmp19*(((tmp44+tmp23+tmp21)*k_jm1_l+(tmp43+tmp23+tmp20)*k_jm1_b)*tmp14+(((tmp44+tmp22+tmp20)*k_jm1_l+(tmp43+tmp22+tmp21)*k_jm1_b)*k_j_b+(tmp42+tmp23+tmp20)*tmp5+(tmp41+tmp23+tmp21)*k_jm1_b*k_jm1_l)*k_j_u+((tmp42+tmp22+tmp21)*tmp5+(tmp41+tmp22+tmp20)*k_jm1_b*k_jm1_l)*k_j_b))/8

    X[1,0] = -(tmp39*tmp17*tmp18*tmp19*(((tmp60+tmp57)*k_jm1_l+(tmp59+tmp58)*k_jm1_b)*tmp14+(((tmp60+tmp48)*k_jm1_l+(tmp59+tmp54)*k_jm1_b)*k_j_b+(tmp56+tmp58)*tmp5+(tmp52+tmp57)*k_jm1_b*k_jm1_l)*k_j_u+((tmp56+tmp54)*tmp5+(tmp52+tmp48)*k_jm1_b*k_jm1_l)*k_j_b))/8

    X[1,1] = (tmp16*tmp17*tmp18*tmp19*((tmp64*k_jm1_l+tmp66*k_jm1_b)*tmp14+((tmp67*k_jm1_l+tmp68*k_jm1_b)*k_j_b+tmp68*tmp5+tmp67*k_jm1_b*k_jm1_l)*k_j_u+(tmp66*tmp5+tmp64*k_jm1_b*k_jm1_l)*k_j_b))/8

    return X



'''Library that implements an effective density model for composition 
profiles (as described by Tolan) and also a home constructed model for
magnetic moment profiles (element specific). 
'''
#from pylab import *
from numpy import *

import numpy as np

from scipy.special import erf

def erf_profile(z, z0, d, sigma0, sigma1):
    eta = (sigma1*z0 + sigma0*(z0 + d))/(sigma0 + sigma1)
    p = 0.5*(1 + where(z <= eta, erf((z - z0)/sqrt(2.)/sigma0), 
                       -erf((z - z0 - d)/sqrt(2.)/sigma1)))
    return p

def erf_interf(z, sigma):
    return 0.5 + 0.5*erf(z/sqrt(2.)/sigma)

def _lin(z, sigma):
    p = zeros(z.shape)
    p = ((z > sqrt(3)*sigma)*1.0 + 
         (abs(z) <= sqrt(3)*sigma)*(0.5 + z/2/sqrt(3)/sigma))
    return p

def lin_profile(z, z0, d, sigma0, sigma1):
    eta = (sigma1*z0 + sigma0*(z0 + d))/(sigma0 + sigma1)
    p =  where(z < eta, _lin(z - z0, sigma0), 
                       _lin(-(z - z0 - d), sigma1))
    return p

def _exp(z, sigma):
    p = ((z <= 0)*(0.5*exp(sqrt(2)*z/sigma)) + 
         (z > 0)*(1 - 0.5*exp(-sqrt(2)*z/sigma)))
    return p

def exp_profile(z, z0, d, sigma0, sigma1):
    eta = (sigma1*z0 + sigma0*(z0 + d))/(sigma0 + sigma1)
    p =  where(z < eta, _exp(z - z0, sigma0), 
                       _exp(-(z - z0 - d), sigma1))
    return p

def _sin(z, sigma):
    a = pi/sqrt(pi**2 - 8)*sigma
    p = (z > a)*1.0 + (abs(z) <= a)*(0.5 + 0.5*sin(pi*z/2/a))
    return p

def sin_profile(z, z0, d, sigma0, sigma1):
    eta = (sigma1*z0 + sigma0*(z0 + d))/(sigma0 + sigma1)
    p =  where(z < eta, _sin(z - z0, sigma0), 
                       _sin(-(z - z0 - d), sigma1))
    return p


def compress_profile_old(z, p, delta_max):
    pnew = p.copy()
    znew = z.copy()
    i = 0
    index = array([False])
    while any(bitwise_not(index)):
        #print i
        i += 1
        index = abs(pnew[:-1] - pnew[1:]) > delta_max
        index[::2] = True
        index = r_[index, True]
        #print pnew.shape, index.shape
        pnew = pnew[index]
        znew = znew[index]

    return znew, pnew

def compress_profile(z, p, delta_max):
    ''' Compresses a profile by merging the neighbouring layers that
    has an density difference of maximum delta_max'''
    pnew = p.copy()
    znew = z.copy()
    inew = arange(len(p))
    i = 0
    index = array([False])
    while any(bitwise_not(index)):
        #print i
        i += 1
        index = array([True]*len(pnew))
        index[1:-1:2] = abs(pnew[:-2:2] - pnew[2::2]) > delta_max
        #print pnew.shape, index.shape
        pnew = pnew[index]
        znew = znew[index]
        inew = inew[index]
    #print inew
    inew = append(inew, inew[-1])
    pret = array([p[inew[i]:inew[i+1]+1].mean() for i in range(len(inew)-1)])
    return znew, pret

def compress_profile2(z, p1, p2, delta_max):
    ''' Compresses a profile by merging the neighbouring layers that
    has an density difference of maximum delta_max'''
    p1new = p1.copy()
    p2new = p2.copy()
    znew = z.copy()
    inew = arange(len(p1))
    i = 0
    index = array([False])
    while any(bitwise_not(index)):
        #print i
        i += 1
        index = array([True]*len(p1new))
        index[1:-1:2] = bitwise_or(abs(p1new[:-2:2] - p1new[2::2]) > delta_max
                                    ,abs(p2new[:-2:2] - p2new[2::2]) > delta_max
                                    )
        #print pnew.shape, index.shape
        p1new = p1new[index]
        p2new = p2new[index]
        znew = znew[index]
        inew = inew[index]
    #print inew
    inew = append(inew, inew[-1])
    p1ret = array([p1[inew[i]:inew[i+1]+1].mean() for i in range(len(inew)-1)])
    p2ret = array([p2[inew[i]:inew[i+1]+1].mean() for i in range(len(inew)-1)])
    return znew, p1ret, p2ret


def compress_profile_n(z, ps, delta_maxs):
    ''' Compresses multiple profiles, each profile: ps[i], 
    by merging the neighbouring layers that
    has an density difference of maximum delta_maxs[i]'''
    inew, znew = compress_profile_index_n(z, ps, delta_maxs)
    psnew = create_compressed_profile(ps, inew)
    return znew, psnew


def compare_layers(ps, delta_max):
    """Compare consecutive layers and see if ps is smaller than delta_max"""
    if len(ps.shape) > 1:
        return (abs(ps[:-2:2] - ps[2::2]) > delta_max).any(axis=1)
    else:
        return abs(ps[:-2:2] - ps[2::2]) > delta_max


def compress_profile_index_n(z, ps, delta_maxs):
    znew = z.copy()
    inew = arange(len(ps[0]))
    psnew = [ps_i.copy() for ps_i in ps]
    i = 0
    index = array([False])
    while any(bitwise_not(index)):
        i += 1
        index = array([True]*len(psnew[0]))
        test = array([compare_layers(ps_i, delta_maxs_i) for ps_i, delta_maxs_i in zip(psnew, delta_maxs)])
        index[1:-1:2] = test.any(axis=0)
        
        psnew = [ps_i[index] for ps_i in psnew]
        znew = znew[index]
        inew = inew[index]
    inew = append(inew, inew[-1])
    return inew, znew


def create_compressed_profile(ps, inew):
    psret = [array([ps_i[inew[i]:inew[i+1]+1].mean(0) for i in range(len(inew)-1)]) for ps_i in ps]
    return psret


def create_profile(d, sigma, dens, prof_funcs, dz = 0.01,
                   mult = 3, buffer = 0, delta = 20):
    zlay = r_[0, cumsum(d)]
    z = arange(-sigma[0]*mult - buffer, zlay[-1]+sigma[-1]*mult + buffer, dz)
    ptot = z*0
    pdens = z*0
    ps = []
    d = r_[delta+sigma[0]*mult + buffer, d, sigma[-1]*mult+delta + buffer]
    sigma = r_[0, sigma, 0]
    zlay = cumsum(d)
    #print zlay
    z0 = buffer + sigma[1]*mult
    zlay = r_[0, zlay] - z0
    #print zlay
    #for i in range(len(d)):
    #    print z0
    #    p = make_profile(z, z0, d[i], sigma[i], sigma[i+1])
    #    z0 += d[i]
    #    ptot += p
    #    pdens += p*dens[i]
    #    ps.append(p)
    ps = array([prof(z, zp, di, s_low, s_up) 
                for prof, zp, di, s_low, s_up in zip(prof_funcs, zlay, 
                                                     d, sigma[:-1], sigma[1:])])
    ptot = ps.sum(0)
    pdens_indiv = (ps*dens[:,newaxis])/ptot
    pdens = pdens_indiv.sum(0)

    return z, pdens, pdens_indiv


def create_profile_cm(d, sigma_c, sigma_m, prof_funcs,
                      prof_funcs_mag, dmag_dens_l, dmag_dens_u, mag_dens, dd_m,
                      dz = 0.01, mult = 3, buffer = 0, delta = 20):
    '''Create a scattering length profile for magnetic x-ray scattering for the charge and
    magnetic part of the scattering lengths. sigma is roughnesses.
    Note that prof_funcs relate to layer profiles and prof_funcs_mag relates to 
    magnetic interface profiles.
    '''
    zlay = r_[0, cumsum(d)]
    s_max_bot = max(sigma_c[0], sigma_m[0])
    s_max_up = max(sigma_c[-1], sigma_m[-1])
    z = arange(-s_max_bot*mult - buffer, zlay[-1]+s_max_up*mult + buffer + dz, dz)
    ptot = z*0
    pdens = z*0
    ps = []
    d = r_[delta+s_max_bot*mult + buffer, d, s_max_up*mult+delta + buffer]
    sigma_c = r_[0, sigma_c, 0] + 1e-20
    sigma_m = r_[0, sigma_m, 0] + 1e-20
    dd_m = r_[0, dd_m, 0] + 1e-20
    prof_funcs_mag = [prof_funcs_mag[0]] + prof_funcs_mag + [prof_funcs_mag[-1]]
    zlay = cumsum(d)
    z0 = delta + buffer + s_max_bot*mult
    zlay = r_[0, zlay] - z0
    ps = array([r_[prof(z, zp, di, s_c_low, s_c_up), 
                   #prof(z, zp, di, s_m_low, s_m_up)]
                   prof_m_l(-(z - zp - dd_low), s_m_low)*dm_l + prof_m_u(z - zp - di - dd_up, s_m_up)*dm_u + m]
                for prof, zp, di, s_c_low, s_c_up, s_m_low, s_m_up,
                prof_m_l, prof_m_u, dm_l, dm_u, m, dd_low, dd_up
                   in zip(prof_funcs, zlay, d, sigma_c[:-1], sigma_c[1:], 
                          sigma_m[:-1], sigma_m[1:],
                          prof_funcs_mag[:-1], prof_funcs_mag[1:],
                          dmag_dens_l, dmag_dens_u, mag_dens, dd_m[:-1], dd_m[1:]
                          )])
    ptot = ps.sum(0)
    #print ps.shape, ptot.shape
    pdens_indiv = (ps[:, :len(z)])/ptot[:len(z)]
    pdens_indiv_c = pdens_indiv#*dens_c[:,newaxis]
    pdens_indiv_m = ps[:, len(z):]#*dens_m[:,newaxis]
    #pdens_c = pdens_indiv_c.sum(0)
    #pdens_m = (pdens_indiv_m*pdens_indiv).sum(0)

    return (z, pdens_indiv_c, pdens_indiv_m)


def create_profile_cm2(d, sigma_c, sigma_ml, sigma_mu, prof_funcs,
                      prof_funcs_mag, dmag_dens_l, dmag_dens_u, mag_dens, dd_l, dd_u,
                      dz = 0.01, mult = 3, buffer = 0, delta = 20):
    '''Create a scattering length profile for magnetic x-ray scattering for the charge and
    magnetic part of the scattering lengths. sigma is roughnesses.
    Note that prof_funcs relate to layer profiles and prof_funcs_mag relates to 
    magnetic interface profiles.
    '''
    zlay = r_[0, cumsum(d)]
    s_max_bot = max(sigma_c[0], sigma_ml[0])
    s_max_up = max(sigma_c[-1], sigma_mu[-1])
    z = arange(-s_max_bot*mult - buffer, zlay[-1]+s_max_up*mult + buffer + dz, dz)
    ptot = z*0
    pdens = z*0
    ps = []
    d = r_[delta+s_max_bot*mult + buffer, d, s_max_up*mult+delta + buffer]
    sigma_c = r_[0, sigma_c, 0] + 1e-20
    sigma_mu = r_[sigma_mu, 0] + 1e-20
    sigma_ml = r_[0,sigma_ml] + 1e-20
    
    dd_l = r_[dd_l, 0] + 1e-20
    dd_u = r_[dd_u, 0] + 1e-20
    
    prof_funcs_mag = [prof_funcs_mag[0]] + prof_funcs_mag + [prof_funcs_mag[-1]]
    zlay = cumsum(d)
    z0 = delta + buffer + s_max_bot*mult
    zlay = r_[0, zlay] - z0
    ps = array([r_[prof(z, zp, di, s_c_low, s_c_up), 
                   #prof(z, zp, di, s_m_low, s_m_up)]
                   prof_m_l(-(z - zp - dd_low), s_m_low)*dm_l*m + prof_m_u((z - zp - di + dd_up), s_m_up)*dm_u*m + m]
                for prof, zp, di, s_c_low, s_c_up, s_m_low, s_m_up,
                prof_m_l, prof_m_u, dm_l, dm_u, m, dd_low, dd_up
                   in zip(prof_funcs, zlay, d, sigma_c[:-1], sigma_c[1:], 
                          sigma_ml, sigma_mu,
                          prof_funcs_mag[:-1], prof_funcs_mag[1:],
                          dmag_dens_l, dmag_dens_u, mag_dens, dd_l, dd_u
                          )])
    ptot = ps.sum(0)
    #print ps.shape, ptot.shape
    pdens_indiv = (ps[:, :len(z)])/ptot[:len(z)]
    pdens_indiv_c = pdens_indiv#*dens_c[:,newaxis]
    pdens_indiv_m = ps[:, len(z):]#*dens_m[:,newaxis]
    #pdens_c = pdens_indiv_c.sum(0)
    #pdens_m = (pdens_indiv_m*pdens_indiv).sum(0)

    return (z, pdens_indiv_c, pdens_indiv_m)

if __name__ == '__main__':
    from pylab import *
    from mpl_toolkits.axes_grid import Grid
    from mpl_toolkits.axes_grid.anchored_artists import AnchoredText
    #sigma0 = 5.
    #sigma1 = 1.
    #mult = 3
    #buffer = 0
    #delta = 20
    #z0 = sigma0*mult
    #d = 9.#sigma0*mult + sigma1*mult
    #z = arange(0, z0 + d + sigma1*mult, 0.1)
    #p = make_profile(z, z0, d, sigma0, sigma1)
    #plot(z,p)
    sigma_c = array([4,4, 2])*1.0#sigma_c = array([2,8,2])*1.0
    sigma_m = array([1,1,1])*1.0#sigma_m = array([2,5,4])*1.0
    dmag_l = array([0.0,0.0,0.0])
    dmag_u = array([0.0, 0.0, 0.0])
    mag_dens = array([0.0, 2.25, 0.0])
    #ddm = array([10.0, -10.0, 0.0])
    ddm = array([10.0, -10.0, 10.0])
    d = array([0.0, 50, 0.0])*1.0#d = array([5, 60])*1.0
    dens_c = array([-1.79e-3 + 0.0029J, -4e-23])#dens_c = array([5,50,25,0])
    dens_m = array([1.4458e-3 - 0.00144J, 2.12e-23])#dens_m = array([0,6,10,0])
    prof_funcs = [erf_profile]*3#prof_funcs = [erf_profile]*4
    mag_prof_funcs = [erf_interf]*3
    #z, pdens, pdens_indiv = create_profile(d, sigma_c, dens_cprof_funcs, dz = 0.1,
    #                                       mult = 5.0, 
    #                                       buffer = 0, delta = 5.)
    z, c_den, m_den = create_profile_cm(d[1:-1], sigma_c[:-1], sigma_m[:-1],  
                                        prof_funcs, mag_prof_funcs, 
                                        dmag_l,dmag_u, mag_dens,
                                        ddm, dz = 1.0, mult = 4.0, 
                                        buffer = 20, delta = 5.)
    def plot_fig(filename, fig_title = ''):
        fig = figure(1)
        grid = Grid(fig, 111, nrows_ncols=(3,1), axes_pad = 0.15)
        #subplot(311)
        grid[0].set_title(fig_title)    
        grid[0].plot(z, c_den[0], lw = 2.0, c = 'b')
        grid[0].plot(z, c_den[1], lw = 2.0, c = 'r')
        grid[0].plot(z, c_den[2], lw = 2.0, c = 'g')
        grid[0].axvline(0, linestyle = '--', c = 'k')
        grid[0].axvline(d[1], linestyle = '--', c = 'k')
        grid[0].set_ylabel('Comp.')
        grid[0].legend(('Substrate', 'Fe layer', 'Ambient'))
        #subplot(312)
        grid[1].plot(z, m_den[1], lw = 2.0, c = 'r')
        grid[1].axvline(0, linestyle = '--', c = 'k')
        grid[1].axvline(d[1], linestyle = '--', c = 'k')
        grid[1].axvline(0 + ddm[0], linestyle = ':', c = 'k')
        grid[1].axvline(d[1]+ddm[1], linestyle = ':', c = 'k')
        #grid[1].annotate('Sub.dd_m = %.1f'%ddm[0], xy=(ddm[0], mag_dens[1]), xycoords = 'data',
        #                 xytext = (-50, 0), textcoords = 'offset points', 
        #                 arrowprops=dict(arrowstyle="->", connectionstyle = 'arc3'))
        #grid[1].annotate('Fe.dd_m = %.1f'%ddm[1], xy=(d[1]+ddm[1], mag_dens[1]), xycoords = 'data',
        #                 xytext = (50, 0), textcoords = 'offset points', 
        #                 arrowprops=dict(arrowstyle="<-", connectionstyle = 'arc3'))
        at = AnchoredText("dd_ml = %.1f"%(ddm[0]),
                  prop=dict(size=10), frameon=True,
                  loc=6,
                  )
        at.patch.set_boxstyle("round,pad=0.,rounding_size=0.2")
        grid[1].add_artist(at)
        at = AnchoredText("dd_mu = %.1f"%(ddm[1]),
                  prop=dict(size=10), frameon=True,
                  loc=7,
                  )
        at.patch.set_boxstyle("round,pad=0.,rounding_size=0.2")
        grid[1].add_artist(at)
        grid[1].set_ylabel('Mag. mom.\n per res. atom.')
        grid[1].set_ylim(0, 3.1)
        #subplot(313)
        grid[2].plot(z, m_den[1]*c_den[1], 'r', z, 2.25*c_den[1], 'b', lw = 2.0)
        grid[2].axvline(0, linestyle = '--', c = 'k')
        grid[2].axvline(d[1], linestyle = '--', c = 'k')
        grid[2].axvline(0 + ddm[0], linestyle = ':', c = 'k')
        grid[2].axvline(d[1]+ddm[1], linestyle = ':', c = 'k')
        grid[2].set_ylabel('Tot. Mag mom.')
        grid[2].legend(('Profile', 'Bulk Fe'))
        grid[2].set_xlabel('Depth [AA]')
        savefig(filename)
        clf()
    
    plot_fig('Bulk_Fe_sigma', 'Bulk Fe with no change at the interfaces')
    dmag_l = array([0.0,0.75,0.0])
    dmag_u = array([0.0, 0.75, 0.0])
    z, c_den, m_den = create_profile_cm(d[1:-1], sigma_c[:-1], sigma_m[:-1],  
                                        prof_funcs, mag_prof_funcs, 
                                        dmag_l,dmag_u, mag_dens,
                                        ddm, dz = 1.0, mult = 4.0, 
                                        buffer = 20, delta = 5.)
    plot_fig('Enhanced_Fe_sigma', 'Bulk Fe with enhanced interfaces')
    dmag_l = array([0.0,-1.,0.0])
    dmag_u = array([0.0, -1., 0.0])
    z, c_den, m_den = create_profile_cm(d[1:-1], sigma_c[:-1], sigma_m[:-1],  
                                        prof_funcs, mag_prof_funcs, 
                                        dmag_l,dmag_u, mag_dens,
                                        ddm, dz = 1.0, mult = 4.0, 
                                        buffer = 20, delta = 5.)
    plot_fig('Depleted_Fe_sigma', 'Bulk Fe with depleted interfaces')
        


'''
=====================================================
:mod:`sns_mr` SNS magnetism reflectometer data loader
=====================================================

Loads the default datafile format used for extracted reflectivity at
the SNS magnetism reflectometer. It allows to import several channels at
once, automatically naming the datasets according to the imported
polarizations.
'''

import numpy as np
import wx

from ..data_loader_framework import Template
from ..utils import ShowWarningDialog

class Plugin(Template):
    wildcard = '*.dat'

    def __init__(self, parent):
        Template.__init__(self, parent)
        self.x_col = 0
        self.y_col = 1
        self.e_col = 2
        self.xe_col = 3
        self.ai_col = 4
        self.comment = '#'
        self.skip_rows = 0
        self.delimiter = None

    def LoadData(self, data_item_number, filename):
        '''LoadData(self, data_item_number, filename) --> none

        Loads the data from filename into the data_item_number.
        '''
        # get scan number and polarization channel from filename
        name = ''
        fhandle = open(filename, 'r', encoding='utf-8')
        fline = fhandle.readline()
        while not '[Data]' in fline:
            if 'Input file indices:' in fline:
                name += fline.split(':')[1].strip()
            if 'Extracted states:' in fline:
                name += ' (%s)'%(fline.split(':')[1].strip().split(' ')[0])
            fline = fhandle.readline()
            if not fline:
                ShowWarningDialog(self.parent, 'Could not load the file: ' + filename+' \nWrong format.\n')
                return

        try:
            load_array = np.loadtxt(fhandle, delimiter=self.delimiter, comments=self.comment, skiprows=self.skip_rows)
        except Exception as e:
            ShowWarningDialog(self.parent, 'Could not load the file: ' + filename +
                              ' \nPlease check the format.\n\n numpy.loadtxt' + ' gave the following error:\n' + str(e))
            return
        else:
            # For the freak case of only one data point
            if len(load_array.shape) < 2:
                load_array = np.array([load_array])
            # Check so we have enough columns
            if load_array.shape[1] - 1 < max(self.x_col, self.y_col, self.e_col):
                ShowWarningDialog(self.parent, 'The data file does not contain' + 'enough number of columns. It has '
                                  + str(load_array[1]) + ' columns. Rember that the column index start at zero!')
                # Okay now we have showed a dialog lets bail out ...
                return
            # The data is set by the default Template.__init__ function, neat hu
            # Know the loaded data goes into *_raw so that they are not
            # changed by the transforms
            self.data[data_item_number].x_raw=load_array[:, self.x_col]
            self.data[data_item_number].y_raw=load_array[:, self.y_col]
            self.data[data_item_number].error_raw=load_array[:, self.e_col]
            self.data[data_item_number].set_extra_data('res', load_array[:,self.xe_col], 'res')
            self.data[data_item_number].set_extra_data('ai', load_array[:,self.ai_col], 'ai')
            # Name the dataset accordign to file name
            self.data[data_item_number].name = name
            # Run the commands on the data - this also sets the x,y, error memebers
            # of that data item.
            self.data[data_item_number].run_command()

            # Send an update that new data has been loaded
            self.SendUpdateDataEvent()
            self.UpdateDataList()

    def LoadDataFile(self, selected_items):
        '''LoadDataFile(self, seleceted_items) --> None

        Selected items is the selcted items in the items in the current DataList
        into which data from file(s) should be loaded. Note that the default
        implementation only allows the loading of a single file!
        Overriding this function in subclasses can of course change this
        behaviour. This function calls the LoadData function which implements
        the io function by it self. The LoadData has to be overloaded in
        order to have a working plugin.
        '''
        n_selected = len(selected_items)
        dlg = wx.FileDialog(self.parent, message="Choose your Datafile", wildcard="All files (%s)|%s" % (self.wildcard,
                                                                                                         self.wildcard)
                           , style=wx.FD_OPEN|wx.FD_MULTIPLE|wx.FD_CHANGE_DIR)

        if dlg.ShowModal() == wx.ID_OK:
            files = dlg.GetPaths()
            dlg.Destroy()
            if len(files) != n_selected:
                dlg = wx.MessageDialog(self.parent, 'Please select %i dataset(s)' % n_selected,
                                       caption='Wrong number of selections', style=wx.OK | wx.ICON_INFORMATION)
                dlg.ShowModal()
                dlg.Destroy()
                return self.LoadDataFile(selected_items)
            for i, fname in enumerate(files):
                self.LoadData(selected_items[i], fname)
            return True
        return False



'''
====================================
:mod:`sls_sxrd` SLD sxrd data loader
====================================

Loads the data from whitespace seperated column formatted ascii data files.
It is intended for surface x-ray diffraction data where the data sets consists
of rod scans along the l-direction (perpendicular to the surface). The plugin
sorts each rod with equal h and k values into one data sets. The l-direction 
is also sorted.

The default columns are the following:

* First column h values
* Second column k values
* Third values l values
* Fourth column Intensites
* Fifth column The standard deviation of the intensities.

The other settings are just as in the default data loader.

The h,k values is stored as extra data in data.extra_data dictonary as
h and k. 
'''

import numpy as np
import wx
from wx.lib.masked import NumCtrl

from ..data_loader_framework import Template
from ..utils import ShowErrorDialog, ShowWarningDialog, ShowInfoDialog

class Plugin(Template):
    def __init__(self, parent):
        Template.__init__(self, parent)
        self.h_col = 0
        self.k_col = 1
        self.l_col = 2
        self.I_col = 3
        self.eI_col = 4
        self.comment = '#'
        self.skip_rows = 0
        self.delimiter = None
    
    def LoadData(self, data_item_number, filename):
        '''LoadData(self, data_item_number, filename) --> none
        
        Loads the data from filename into the data_item_number.
        '''
        try:
            load_array = np.loadtxt(filename, delimiter = self.delimiter, 
                comments = self.comment, skiprows = self.skip_rows)
        except Exception as e:
            ShowWarningDialog(self.parent, 'Could not load the file: ' +\
                    filename + ' \nPlease check the format.\n\n numpy.loadtxt'\
                    + ' gave the following error:\n'  +  str(e))
        else:
            # For the freak case of only one data point
            if len(load_array.shape) < 2:
                load_array = np.array([load_array])
            # Check so we have enough columns
            if load_array.shape[1]-1 < max(self.h_col, self.k_col, self.l_col,\
                    self.I_col, self.eI_col):
                ShowWarningDialog(self.parent, 'The data file does not contain'\
                        + 'enough number of columns. It has ' + str(load_array.shape[1])\
                        + ' columns. Rember that the column index start at zero!')
                # Okay now we have showed a dialog lets bail out ...
                return
            # The data is set by the default Template.__init__ function, neat hu
            # Note that the loaded data goes into *_raw so that they are not
            # changed by the transforms
            
            # Create an record array so we can sort the data properly
            data = np.rec.fromarrays([\
                     load_array[:,self.h_col].round().astype(type(1)),\
                     load_array[:,self.k_col].round().astype(type(1)),\
                     load_array[:,self.l_col], load_array[:,self.I_col],\
                     load_array[:,self.eI_col]\
                    ],\
                     names = 'h, k, l, I, eI')
            # Sort the data
            data.sort(order = ('h','k','l'))
            i = 0
            while i < len(data):
                # Find all the data for each rod
                tmp = data.compress(np.bitwise_and(data['h'] == data[i]['h'],\
                                    data['k'] == data[i]['k']))
                self.data.add_new('(%i, %i)'%(tmp['h'][0], tmp['k'][0]))
                self.data[-1].x_raw = tmp['l']
                self.data[-1].y_raw =tmp['I']
                self.data[-1].error_raw = tmp['eI']
                # Run the commands on the data - this also sets the x,y, error memebers
                # of that data item.
                self.data[-1].run_command()
                self.data[-1].set_extra_data('h', tmp['h'], 'h')
                self.data[-1].set_extra_data('k', tmp['k'], 'k')
                # Increase the index
                i += len(tmp)
            
            # Update the data list
            self.UpdateDataList()
            # Send an update that new data has been loaded
            self.SendUpdateDataEvent()
            
        
    def SettingsDialog(self):
        '''SettingsDialog(self) --> None
        
        This function should - if necessary implement a dialog box
        that allows the user set import settings for example.
        '''
        col_values = {'I': self.I_col,'h': self.h_col,'k': self.k_col,\
                        'l': self.l_col, 'I error': self.eI_col}
        misc_values = {'Comment': str(self.comment), 'Skip rows': self.skip_rows,\
                'Delimiter': str(self.delimiter)}
        dlg = SettingsDialog(self.parent, col_values, misc_values)
        if dlg.ShowModal() == wx.ID_OK:
            col_values = dlg.GetColumnValues()
            misc_values = dlg.GetMiscValues()
            self.h_col = col_values['h']
            self.k_col = col_values['k']
            self.l_col = col_values['l']
            self.I_col = col_values['I']
            self.eI_col = col_values['I error']
            self.comment = misc_values['Comment']
            self.skip_rows = misc_values['Skip rows']
            self.delimiter = misc_values['Delimiter']
            self.SetStatusText('Changed import settings')
        else:
            self.SetStatusText('No change to import settings')
        dlg.Destroy()
        
        
class SettingsDialog(wx.Dialog):
    
    def __init__(self, parent, col_values, misc_values):
        wx.Dialog.__init__(self, parent, -1, 'Data loader settings')
        
        box_sizer = wx.BoxSizer(wx.HORIZONTAL)
        
        # Make the box for putting in the columns
        col_box = wx.StaticBox(self, -1, "Columns" )
        col_box_sizer = wx.StaticBoxSizer(col_box, wx.VERTICAL )
        
        
        #col_values = {'y': 1,'x': 0,'y error': 1}
        col_grid = wx.GridBagSizer(len(col_values), 2)
        self.col_controls = col_values.copy()
        keys = list(col_values.keys())
        keys.sort()
        for i, name in enumerate(keys):
            text = wx.StaticText(self, -1, name+': ')
            control = wx.SpinCtrl(self)
            control.SetRange(0,100)
            control.SetValue(col_values[name])
            col_grid.Add(text, (i,0),\
                    flag = wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL,\
                    border = 5)
            col_grid.Add(control, (i,1),\
                    flag = wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL,\
                    border = 5)
            self.col_controls[name] = control
        
        col_box_sizer.Add(col_grid, 0, wx.ALIGN_CENTRE|wx.ALL, 5)
        box_sizer.Add(col_box_sizer, 0, wx.ALL|wx.EXPAND, 5)
        
        col_box = wx.StaticBox(self, -1, "Misc" )
        col_box_sizer = wx.StaticBoxSizer(col_box, wx.VERTICAL )
        
        # Lets add another box for comments and rows to skip
        #misc_values = {'Comment': '#', 'Skip rows': 0,'Delimiter': 'None'}
        col_grid = wx.GridBagSizer(len(misc_values), 2)
        self.misc_controls = misc_values.copy()
        keys = list(misc_values.keys())
        keys.sort()
        for i, name in enumerate(keys):
            text = wx.StaticText(self, -1, name+': ')
            if type(misc_values[name]) == type(1):
                control = wx.SpinCtrl(self)
                control.SetRange(0,100)
                control.SetValue(misc_values[name])
            else:
                control = wx.TextCtrl(self, value = misc_values[name],\
                        style = wx.EXPAND)
            col_grid.Add(text, (i,0),\
                    flag = wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL,\
                    border = 5)
            col_grid.Add(control, (i,1),\
                    flag = wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL,\
                    border = 5)
            self.misc_controls[name] = control
        
        col_box_sizer.Add(col_grid, 0, wx.ALIGN_CENTRE|wx.ALL, 5)
        box_sizer.Add(col_box_sizer, 0, wx.ALL|wx.EXPAND, 5)
        
        button_sizer = wx.StdDialogButtonSizer()
        okay_button = wx.Button(self, wx.ID_OK)
        okay_button.SetDefault()
        button_sizer.AddButton(okay_button)
        button_sizer.AddButton(wx.Button(self, wx.ID_CANCEL))
        button_sizer.Realize()
        
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(box_sizer, 1, wx.GROW, 20)
        line = wx.StaticLine(self, -1, size=(20,-1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW, 30)
        
        sizer.Add(button_sizer,0,\
                flag = wx.ALIGN_RIGHT, border = 20)
        self.SetSizer(sizer)
        
        sizer.Fit(self)
        self.Layout()
        
    def GetColumnValues(self):
        values = {}
        for key in self.col_controls:
            values[key] = self.col_controls[key].GetValue()
        return values
    
    def GetMiscValues(self):
        values = {}
        for key in self.misc_controls:
            val = self.misc_controls[key].GetValue()
            if (type(val) == type('') or type(val) == type('')):
                if val.lower() == 'none':
                    val = None
            values[key] = val
        return values
